// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cstdlib>
#include <iostream>
#include <algorithm>
#include <SDL.h>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "util/Real.hpp"
#include "Config.hpp"
#include "Main.hpp"
#include "Audio.hpp"

#if !defined(AUDIO_FILES_PREFIX)
# define AUDIO_FILES_PREFIX FILES_DIRECTORY
#endif // not defined AUDIO_FILES_PREFIX

#define SPEC_FREQ     44100
#define SPEC_FORMAT   AUDIO_S16SYS
#define SPEC_CHANNELS 2
#define SPEC_SAMPLES  4096

// the module has been initialized
bool Audio::initialized = false;

// all sounds data
Mix_Chunk *Audio::sound_data[SOUND_AMOUNT];

// number of active redholes
int Audio::activeRedHole;
// true if sound is playing
bool Audio::playing[SOUND_AMOUNT];

// initializes the audio [depends on: Config]
void Audio::init(){
	if ( Mix_OpenAudio( SPEC_FREQ, SPEC_FORMAT, SPEC_CHANNELS, SPEC_SAMPLES ) < 0 ){
		std::cerr << "FATAL: cannot start audio: " << Mix_GetError() << std::endl;
		Main::quit( Main::ERROR_AUDIO );
	}
	initialized = true;

	if ( Mix_AllocateChannels( SOUND_AMOUNT ) != SOUND_AMOUNT ){
		std::cerr << "FATAL: not enough memory to start audio" << std::endl;
		Main::quit( Main::ERROR_AUDIO2 );
	}

	loadSounds();

	activeRedHole = 0;
	foreach( bool& p, playing )
		p = false;

	Mix_ChannelFinished( &audiocallback );

	reinit();
}
// reinitializes the audio, call it when config changes
void Audio::reinit(){
	if ( ! Config::get_audio() )
		Mix_Pause( -1 );
	else{
		Mix_Resume( -1 );
		Mix_Volume( -1, static_cast<int>((Config::get_audio()/static_cast<real>(MAX_VOLUME))*MIX_MAX_VOLUME) );
	}
}
// ends the audio
void Audio::finalize(){
	if ( ! initialized )
		return;

	Mix_ChannelFinished( NULL );

	Mix_HaltChannel( -1 );

	foreach( Mix_Chunk *c, sound_data ){
		if ( c != NULL )
			Mix_FreeChunk( c );
	}

	Mix_CloseAudio();
}
// stops the audio
void Audio::stop(){
	Mix_Pause( -1 );
}
// audio file loader
void Audio::loadSounds(){
	const char *file_names[] = {
		AUDIO_FILES_PREFIX "bomb.ogg",
		AUDIO_FILES_PREFIX "bullet_fire.ogg",
		AUDIO_FILES_PREFIX "bullet_hit.ogg",
		AUDIO_FILES_PREFIX "enemy_birth.ogg",
		AUDIO_FILES_PREFIX "game_over.ogg",
		AUDIO_FILES_PREFIX "game_start.ogg",
		AUDIO_FILES_PREFIX "laser_fire.ogg",
		AUDIO_FILES_PREFIX "lightning_fire.ogg",
		AUDIO_FILES_PREFIX "menu_accept.ogg",
		AUDIO_FILES_PREFIX "menu_change.ogg",
		AUDIO_FILES_PREFIX "menu_exit.ogg",
		AUDIO_FILES_PREFIX "menu_move.ogg",
		AUDIO_FILES_PREFIX "player_death.ogg",
		AUDIO_FILES_PREFIX "redhole_drag.ogg",
	};

	for( int id=0 ; id<SOUND_AMOUNT ; ++id ){
		sound_data[id] = Mix_LoadWAV( file_names[id] );

		if ( sound_data[id] == NULL )
			std::cerr << "ERROR: sound \"" << file_names[id] << "\" could not be loaded" << std::endl;
	}
}
// plays the sound id
void Audio::playSound( int id, bool silent ){
	if ( ! initialized )
		return;

	if ( id == SOUND_REDHOLE_DRAG )
		++activeRedHole;

	// if it's a redhole loop sound, or audio hasn't been stopped
	if ( (id == SOUND_REDHOLE_DRAG) || Config::get_audio() ){
		if ( !playing[id] && (sound_data[id] != NULL) ){
			playing[id] = true;
			Mix_PlayChannel( id, sound_data[id], 0 );

			// if audio had been stopped, restop it
			if ( silent || !Config::get_audio() )
				Mix_Pause( id );
		}
	}
}
// stops the sound id (to stop the SOUND_REDHOLE_DRAG, stopSound must be called as
// many times as startSound was called)
void Audio::stopSound( int id ){
	if ( ! initialized )
		return;

	if ( id == SOUND_REDHOLE_DRAG ){
		--activeRedHole;
		// redhole loop sound only stops when no redholes are left
		if ( ! activeRedHole )
			Mix_HaltChannel( id );
	}else
		Mix_HaltChannel( id );
}
// callback for SDL_mixer when a channel ends
void Audio::audiocallback( int id ){
	if ( id == SOUND_REDHOLE_DRAG ){
		// checking if it needs to be looped again
		if ( ! activeRedHole )
			playing[id] = false;
		else if ( sound_data[id] != NULL )
			Mix_PlayChannel( id, sound_data[id], 0 );
	}else
		playing[id] = false;
}
