// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cstdlib>
#include <utility>
#include <fstream>
#include <iostream>
#include <algorithm>
#include <SDL.h>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "util/Text.hpp"
#include "Locale.hpp"
#include "Main.hpp"
#include "Config.hpp"

// DEFAULT VALUES

#define DEFAULT_WIDTH       640
#define DEFAULT_HEIGHT      480
#define DEFAULT_BITDEPTH    16
#define DEFAULT_FULLSCREEN  true
#define DEFAULT_LINESMOOTH  false

#define DEFAULT_GRID        true
#define DEFAULT_GRID_ROW    30
#define DEFAULT_GRID_COL    40
#define DEFAULT_GRID_BEZIER false

#define DEFAULT_DEBRIS      true

#define DEFAULT_AUDIO       MAX_VOLUME

#define DEFAULT_HISCORE     0

#define DEFAULT_P1_KESCAPE              SDLK_ESCAPE
#define DEFAULT_P1_KIMM                 SDLK_UNKNOWN
#define DEFAULT_P1_KBOMB                SDLK_SPACE
#define DEFAULT_P1_KUP                  SDLK_UP
#define DEFAULT_P1_KDOWN                SDLK_DOWN
#define DEFAULT_P1_KRIGHT               SDLK_RIGHT
#define DEFAULT_P1_KLEFT                SDLK_LEFT
#define DEFAULT_P1_KFIREUP              SDLK_w
#define DEFAULT_P1_KFIREDOWN            SDLK_s
#define DEFAULT_P1_KFIRERIGHT           SDLK_d
#define DEFAULT_P1_KFIRELEFT            SDLK_a
#define DEFAULT_P1_JESCAPE              std::pair<int,int>( -1, -1 )
#define DEFAULT_P1_JIMM                 std::pair<int,int>( -1, -1 )
#define DEFAULT_P1_JBOMB                std::pair<int,int>( -1, -1 )
#define DEFAULT_P1_JUPDOWN              std::pair<int,int>( -1, -1 )
#define DEFAULT_P1_JINVERTUPDOWN        false
#define DEFAULT_P1_JRIGHTLEFT           std::pair<int,int>( -1, -1 )
#define DEFAULT_P1_JINVERTRIGHTLEFT     false
#define DEFAULT_P1_JFIREUPDOWN          std::pair<int,int>( -1, -1 )
#define DEFAULT_P1_JINVERTFIREUPDOWN    false
#define DEFAULT_P1_JFIRERIGHTLEFT       std::pair<int,int>( -1, -1 )
#define DEFAULT_P1_JINVERTFIRERIGHTLEFT false

// CONFIG FORMAT

#define CONFIG_NAME                  "." PACKAGE_TARNAME ".cfg"

#define CONFIG_SCREEN_CODE           "screen"
#define CONFIG_GRID_CODE             "grid"
#define CONFIG_DEBRIS_CODE           "debris_active"
#define CONFIG_AUDIO_CODE            "audio_volume"
#define CONFIG_HISCORE_CODE          "highest_score"
#define CONFIG_LANG_CODE             "language"
#define CONFIG_PLAYER_CODE           "player_controls"

// config file
std::string Config::configfile;

// screen properties
int  Config::resx;
int  Config::resy;
int  Config::bitdepth;
bool Config::fullscreen;
bool Config::linesmooth;
// grid properties
bool Config::grid;
int  Config::gridrows;
int  Config::gridcols;
bool Config::gridbezier;
// debris
bool Config::debris;
// audio
int Config::audio;
// hiscore
int Config::hiscore;
// language
std::string Config::lang;
// controls
std::vector<Config::playercontrols> Config::controls;

// initializes the Config [depends on: -]
void Config::init(){
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
	configfile = CONFIG_NAME;
#else // Windows
	char *home = std::getenv( "HOME" );
	if ( home != NULL ){
		configfile = home;
		configfile += "/" CONFIG_NAME;
	}else
		std::cerr << "ERROR: cannot get $HOME directory, configuration file won't be used" << std::endl;
#endif // Not Windows

	defaults();
	if ( ! load() )
		defaults();
}
// loads default values
void Config::defaults(){
	const SDL_VideoInfo* vid = SDL_GetVideoInfo();
	// default screen resolution is present resolution, if we discover it
	if ( vid == NULL ){
		resx     = DEFAULT_WIDTH;
		resy     = DEFAULT_HEIGHT;
		bitdepth = DEFAULT_BITDEPTH;
	}else{
		resx     = vid->current_w;
		resy     = vid->current_h;
		bitdepth = vid->vfmt->BitsPerPixel;
	}
	fullscreen = DEFAULT_FULLSCREEN;
	linesmooth = DEFAULT_LINESMOOTH;

	grid       = DEFAULT_GRID;
	gridrows   = DEFAULT_GRID_ROW;
	gridcols   = DEFAULT_GRID_COL;
	gridbezier = DEFAULT_GRID_BEZIER;

	debris = DEFAULT_DEBRIS;

	audio = DEFAULT_AUDIO;

	hiscore = DEFAULT_HISCORE;

	lang = DEFAULT_LANG;

	controls.clear();
	playercontrols plcntrl;
	plcntrl.kescape              = DEFAULT_P1_KESCAPE;
	plcntrl.kimm                 = DEFAULT_P1_KIMM;
	plcntrl.kbomb                = DEFAULT_P1_KBOMB;
	plcntrl.kup                  = DEFAULT_P1_KUP;
	plcntrl.kdown                = DEFAULT_P1_KDOWN;
	plcntrl.kright               = DEFAULT_P1_KRIGHT;
	plcntrl.kleft                = DEFAULT_P1_KLEFT;
	plcntrl.kfireup              = DEFAULT_P1_KFIREUP;
	plcntrl.kfiredown            = DEFAULT_P1_KFIREDOWN;
	plcntrl.kfireright           = DEFAULT_P1_KFIRERIGHT;
	plcntrl.kfireleft            = DEFAULT_P1_KFIRELEFT;
	plcntrl.jescape              = DEFAULT_P1_JESCAPE;
	plcntrl.jimm                 = DEFAULT_P1_JIMM;
	plcntrl.jbomb                = DEFAULT_P1_JBOMB;
	plcntrl.jupdown              = DEFAULT_P1_JUPDOWN;
	plcntrl.jinvertupdown        = DEFAULT_P1_JINVERTUPDOWN;
	plcntrl.jrightleft           = DEFAULT_P1_JRIGHTLEFT;
	plcntrl.jinvertrightleft     = DEFAULT_P1_JINVERTRIGHTLEFT;
	plcntrl.jfireupdown          = DEFAULT_P1_JFIREUPDOWN;
	plcntrl.jinvertfireupdown    = DEFAULT_P1_JINVERTFIREUPDOWN;
	plcntrl.jfirerightleft       = DEFAULT_P1_JFIRERIGHTLEFT;
	plcntrl.jinvertfirerightleft = DEFAULT_P1_JINVERTFIRERIGHTLEFT;
	controls.push_back( plcntrl );
}
// loads saved config, returns false if error while loading.
// it does not erase present config, so stored config can just
// have a couple of values to override part of the defaults
bool Config::load(){
	if ( configfile.empty() )
		return false;

	std::ifstream in( configfile.c_str(), std::ifstream::in );
	if ( ! in ){
		std::cerr << "ERROR: cannot open the configuration file \"" << configfile << "\"" << std::endl;
		return false;
	}

	// no player definied yet, the first definition has to clear player vector
	bool players = false;

	while( in.good() ){
		// getting code
		std::string code;
		std::getline( in, code, '=' );
		if ( ! in ){
			in.clear( in.rdstate() & std::ifstream::badbit );
			break;
		}

		code = util::Text::trim( util::Text::utf8SkipBOM( code ) );
		if ( code.empty() )
			break;

		// screen
		if ( code == CONFIG_SCREEN_CODE ){
			in >> resx >> resy >> bitdepth >> fullscreen >> linesmooth;
			resx     = std::max( 1, resx );
			resy     = std::max( 1, resy );
			bitdepth = std::max( 8, std::min( 32, bitdepth ) );
		// grid
		}else if ( code == CONFIG_GRID_CODE ){
			in >> grid;
			if ( grid ){
				in >> gridrows >> gridcols >> gridbezier;
				gridrows = std::min( ROWS_MAX, std::max( ROWS_MIN, gridrows ) );
				gridcols = std::min( COLS_MAX, std::max( COLS_MIN, gridcols ) );
			}else{
				gridrows = ROWS_MIN;
				gridcols = COLS_MIN;
				gridbezier = DEFAULT_GRID_BEZIER;
			}
		// debris
		}else if ( code == CONFIG_DEBRIS_CODE ){
			in >> debris;
		// audio
		}else if ( code == CONFIG_AUDIO_CODE ){
			in >> audio;
			audio = std::max( 0, std::min( MAX_VOLUME, audio ) );
		// hiscore
		}else if ( code == CONFIG_HISCORE_CODE ){
			in >> hiscore;
			hiscore = std::max( 0, std::min( MAX_SCORE, hiscore ) );
		// language
		}else if ( code == CONFIG_LANG_CODE ){
			std::getline( in, lang );
			lang = util::Text::trim( util::Text::eraseNewLine( lang ) );
		// player
		}else if ( code == CONFIG_PLAYER_CODE ){
			playercontrols plcntrl;

			int  first;
			int  second;
			bool third;

			// kescape
			in >> first;
			plcntrl.kescape = static_cast<SDLKey>(first);
			// kimm
			in >> first;
			plcntrl.kimm = static_cast<SDLKey>(first);
			// kbomb
			in >> first;
			plcntrl.kbomb = static_cast<SDLKey>(first);
			// kup
			in >> first;
			plcntrl.kup = static_cast<SDLKey>(first);
			// kdown
			in >> first;
			plcntrl.kdown = static_cast<SDLKey>(first);
			// kright
			in >> first;
			plcntrl.kright = static_cast<SDLKey>(first);
			// kleft
			in >> first;
			plcntrl.kleft = static_cast<SDLKey>(first);
			// kfireup
			in >> first;
			plcntrl.kfireup = static_cast<SDLKey>(first);
			// kfiredown
			in >> first;
			plcntrl.kfiredown = static_cast<SDLKey>(first);
			// kfireright
			in >> first;
			plcntrl.kfireright = static_cast<SDLKey>(first);
			// kfireleft
			in >> first;
			plcntrl.kfireleft = static_cast<SDLKey>(first);
			// jescape
			in >> first >> second;
			plcntrl.jescape = std::pair<int,int>( first, second );
			// jimm
			in >> first >> second;
			plcntrl.jimm = std::pair<int,int>( first, second );
			// jbomb
			in >> first >> second;
			plcntrl.jbomb = std::pair<int,int>( first, second );
			// jupdown
			in >> first >> second >> third;
			plcntrl.jupdown = std::pair<int,int>( first, second );
			plcntrl.jinvertupdown = third;
			// jrightleft
			in >> first >> second >> third;
			plcntrl.jrightleft = std::pair<int,int>( first, second );
			plcntrl.jinvertrightleft = third;
			// jfireupdown
			in >> first >> second >> third;
			plcntrl.jfireupdown = std::pair<int,int>( first, second );
			plcntrl.jinvertfireupdown = third;
			// jfirerightleft
			in >> first >> second >> third;
			plcntrl.jfirerightleft = std::pair<int,int>( first, second );
			plcntrl.jinvertfirerightleft = third;

			 // the first player controls has to clear player vector
			if ( ! players ){
				controls.clear();
				players = true;
			}
			controls.push_back( plcntrl );
		}else{
			std::cerr << "ERROR: configuration has an unknown code: " << code << std::endl;
			in.close();
			return false;
		}
	}
	if ( ! in ){
		if ( in.bad() )
			std::cerr << "ERROR: I/O error while reading configuration" << std::endl;
		else
			std::cerr << "ERROR: configuration file has bad values" << std::endl;
		in.close();
		return false;
	}

	in.close();
	return true;
}

void Config::save(){
	if ( configfile.empty() )
		return;

	std::ofstream out( configfile.c_str(), std::ofstream::out|std::ofstream::trunc );
	if ( ! out ){
		std::cerr << "ERROR: cannot save the configuration file \"" << configfile << "\"" << std::endl;
		return;
	}

	// screen
	out << CONFIG_SCREEN_CODE << " = " << resx << " " << resy << " " << bitdepth << " " << fullscreen << " " << linesmooth << std::endl;
	// grid
	out << CONFIG_GRID_CODE << " = " << grid;
	if ( grid )
		out << " " << gridrows << " " << gridcols << " " << gridbezier;
	out << std::endl;
	// debris
	out << CONFIG_DEBRIS_CODE << " = " << debris << std::endl;
	// audio
	out << CONFIG_AUDIO_CODE << " = " << audio << std::endl;
	// hiscore
	out << CONFIG_HISCORE_CODE << " = " << Main::get_hiscore() << std::endl;
	// language
	out << CONFIG_LANG_CODE << " = " << lang << std::endl;

	foreach( const playercontrols& p, controls ){
		// player
		out << CONFIG_PLAYER_CODE << " = " << std::endl;

		// kescape
		out << "\011" << static_cast<int>( p.kescape ) << std::endl;
		// kimm
		out << "\011" << static_cast<int>( p.kimm ) << std::endl;
		// kbomb
		out << "\011" << static_cast<int>( p.kbomb ) << std::endl;
		// kup
		out << "\011" << static_cast<int>( p.kup ) << std::endl;
		// kdown
		out << "\011" << static_cast<int>( p.kdown ) << std::endl;
		// kright
		out << "\011" << static_cast<int>( p.kright ) << std::endl;
		// kleft
		out << "\011" << static_cast<int>( p.kleft ) << std::endl;
		// kfireup
		out << "\011" << static_cast<int>( p.kfireup ) << std::endl;
		// kfiredown
		out << "\011" << static_cast<int>( p.kfiredown ) << std::endl;
		// kfireright
		out << "\011" << static_cast<int>( p.kfireright ) << std::endl;
		// kfireleft
		out << "\011" << static_cast<int>( p.kfireleft ) << std::endl;

		// jescape
		out << "\011" << p.jescape.first
		    << " " << p.jescape.second << std::endl;
		// jimm
		out << "\011" << p.jimm.first
		    << " " << p.jimm.second << std::endl;
		// jbomb
		out << "\011" << p.jbomb.first
		    << " " << p.jbomb.second << std::endl;
		// jupdown
		out << "\011" << p.jupdown.first
		    << " " << p.jupdown.second
		    << " " << p.jinvertupdown << std::endl;
		// jrightleft
		out << "\011" << p.jrightleft.first
		    << " " << p.jrightleft.second
		    << " " << p.jinvertrightleft << std::endl;
		// jfireupdown
		out << "\011" << p.jfireupdown.first
		    << " " << p.jfireupdown.second
		    << " " << p.jinvertfireupdown << std::endl;
		// jfirerightleft
		out << "\011" << p.jfirerightleft.first
		    << " " << p.jfirerightleft.second
		    << " " << p.jinvertfirerightleft << std::endl;
	}

	out.close();
}
// getters and setters for screen properties
int  Config::get_resx(){
	return resx;
}
void Config::set_resx( int rx ){
	resx = rx;
}
int  Config::get_resy(){
	return resy;
}
void Config::set_resy( int ry ){
	resy = ry;
}
int  Config::get_bitdepth(){
	return bitdepth;
}
void Config::set_bitdepth( int bd ){
	bitdepth = bd;
}
bool Config::get_fullscreen(){
	return fullscreen;
}
void Config::set_fullscreen( bool f ){
	fullscreen = f;
}
bool Config::get_linesmooth(){
	return linesmooth;
}
void Config::set_linesmooth( bool ls ){
	linesmooth = ls;
}
// getters and setters for grid properties
bool Config::get_grid(){
	return grid;
}
void Config::set_grid( bool g ){
	grid = g;
}
int  Config::get_gridrows(){
	return gridrows;
}
void Config::set_gridrows( int gr ){
	gridrows = gr;
}
int  Config::get_gridcols(){
	return gridcols;
}
void Config::set_gridcols( int gc ){
	gridcols = gc;
}
bool Config::get_gridbezier(){
	return gridbezier;
}
void Config::set_gridbezier( bool gb ){
	gridbezier = gb;
}
// getter and setter for debris
bool Config::get_debris(){
	return debris;
}
void Config::set_debris( bool d ){
	debris = d;
}
// getter and setter for audio
int  Config::get_audio(){
	return audio;
}
void Config::set_audio( int a ){
	audio = a;
}
// getter for hiscore
int Config::get_hiscore(){
	return hiscore;
}
// getter and setter for language
const std::string& Config::get_lang(){
	return lang;
}
void Config::set_lang( const std::string& l ){
	lang = l;
}
// getter and setter for controls
const std::vector<Config::playercontrols>& Config::get_controls(){
	return controls;
}
void Config::set_controls( const std::vector<Config::playercontrols>& c ){
	controls = c;
}
