// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef INPUT_HPP
#define INPUT_HPP

#include <map>
#include <list>
#include <vector>
#include <utility>
#include <SDL.h>
#include "util/Vector.hpp"

class Input{
	public:
		// initializes the input [depends on: Config]
		static void init();
		// reinitializes the input, call it when config changes
		static void reinit();
		// ends the input
		static void finalize();

		// starting a processing round
		static void newround();
		// receiving events
		static void key( SDLKey key, bool pressed );
		static void button( int which, int button, bool pressed );
		static void axis( int which, int axis, int value );

		// gets all input instances
		static const std::vector<Input*>& get_inputs();
		// gets an input instance
		static const Input *get_input( int i );
		// aggregates getters
		static bool get_agg_escape();
		static bool get_agg_imm();
		static bool get_agg_bomb();
		static const util::Vector& get_agg_move();
		static const util::Vector& get_agg_fire();

	private:
		// compute the aggregates
		static void aggregate();
		// creates a vector representation
		static util::Vector createVec( bool up, bool down, bool right, bool left );
		static util::Vector createVec( int updown, int rightleft );

		// key mappings
		static std::map< SDLKey, std::pair<Input*,int> > keymappings; // unordered
		// joystick mappings
		static std::map< std::pair<int,int>, std::pair<Input*,int> > buttonmappings; // unordered
		static std::map< std::pair<int,int>, std::pair<Input*,int> > axismappings; // unordered

		// opened joysticks
		static std::list<SDL_Joystick*> joys;

		// inputs for each player
		static std::vector<Input*> inputs;
		// aggregate input values
		static bool agg_escape;
		static bool agg_imm;
		static bool agg_bomb;
		static util::Vector agg_move;
		static util::Vector agg_fire;

	// non-static part of Input

	public:
		// getters
		bool get_escape() const;
		bool get_imm() const;
		bool get_bomb() const;
		const util::Vector& get_move() const;
		const util::Vector& get_fire() const;

	private:
		// ctor
		Input();
		// dtor
		virtual ~Input();

		// recieve a key event
		void key( int code, bool pressed );
		// recieve joystick events
		void button( int code, bool pressed );
		void axis( int code, int value );

		// personal input values
		bool escape;
		bool imm;
		bool bomb;
		util::Vector move;
		util::Vector fire;

		// personal keys information
		bool up;
		bool down;
		bool right;
		bool left;
		bool fireup;
		bool firedown;
		bool fireright;
		bool fireleft;
		// personal joystick information
		int updown;
		int rightleft;
		int fireupdown;
		int firerightleft;

		// need to invert joystick axis?
		bool jinvertupdown;
		bool jinvertrightleft;
		bool jinvertfireupdown;
		bool jinvertfirerightleft;
};

#endif // Input.hpp
