// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cstdlib>
#include <fstream>
#include <iostream>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#define int_int int,int
#include "au/Automaton.hpp"
#include "au/EnemyScheduler.hpp"
#include "au/Grid.hpp"
#include "gui/GUI.hpp"
#include "gui/Menu.hpp"
#include "util/Random.hpp"
#include "vis/Font.hpp"
#include "vis/Screen.hpp"
#include "Audio.hpp"
#include "Config.hpp"
#include "Locale.hpp"
#include "Input.hpp"
#include "Timer.hpp"
#include "Main.hpp"

#define STARTING_SCORE 0
#define STARTING_LIFES 3
#define STARTING_BOMBS 3
#define MAX_LIFES 3
#define MAX_BOMBS 9

#define SAVEGAME_VERSION 1

// savegame file name
#define SG_NAME "." PACKAGE_TARNAME ".sg"
std::string Main::sg;

// the window is not visible
bool Main::hidden = false;
// taking a screenshot?
bool Main::screenshot;
// enemy cleaning?
bool Main::killenemies;

// copy of rc::powerup_lifes
int Main::powerup_lifes;

// game stats
int  Main::score;
int  Main::hiscore;
bool Main::gameover;
int  Main::lifes;
int  Main::bombs;

// get game stats
int  Main::get_score(){
	return score;
}
int  Main::get_hiscore(){
	return hiscore;
}
bool Main::get_gameover(){
	return gameover;
}
int  Main::get_lifes(){
	return lifes;
}
int  Main::get_bombs(){
	return bombs;
}
bool Main::get_merciless(){
	return !powerup_lifes;
}
// update game stats
void Main::add_score( int scoreit ){
	if ( ! scoreit )
		return;

	int aux = score;

	// updating score
	score = std::min( MAX_SCORE, score+scoreit );
	if ( score > hiscore ){
		// updating hiscore
		hiscore = score;
		gui::GUI::updateHiscore();
	}
	gui::GUI::updateScore();

	// next shot type
	au::Automaton::next_shottype( scoreit );
	// extra life
	if ( powerup_lifes ){
		if ( (aux/powerup_lifes) != (score/powerup_lifes) ){
			if ( lifes < MAX_LIFES ){
				++lifes;
				gui::GUI::updateLifes();
			}
		}
	}
	// extra bomb
	if ( (aux/rc::powerup_bombs()) != (score/rc::powerup_bombs()) ){
		if ( bombs < MAX_BOMBS ){
			++bombs;
			gui::GUI::updateBombs();
		}
	}
	// updating spawning probabilities
	if ( rc::prob_variation_int() > 0 ){
		if ( (aux/rc::prob_variation_int()) != (score/rc::prob_variation_int()) )
			au::EnemyScheduler::update_prob();
	}
}

// quits the game
void Main::quit( int code ){
	if ( code == NORMAL_EXIT )
		Config::save(); // saving config only if it is a normal exit

	au::Automaton::finalize();
	activeplayers.clear();
	rebirthwait.clear();

	Audio::finalize();
	vis::Font::finalize();
	au::Grid::finalize();
	gui::GUI::finalize();
	Input::finalize();
	gui::Menu::finalize();

	Locale::finalize();

	rc::finalize();

	SDL_Quit();
	std::exit( code );
}
// starts game modules in proper order (see their dependencies)
void Main::start_modules(){
	Locale::preinit();

	rc::init();

	au::Automaton::init();
	Config::init();
	util::Random::init();

	Locale::init();

	Audio::init();
	au::EnemyScheduler::init();
	vis::Font::init();
	au::Grid::init();
	gui::GUI::init();
	Input::init();
	gui::Menu::init();

	vis::Screen::init();

	Timer::init();
}
// clears everything
void Main::clear(){
	// removing players
	au::Automaton::kill_players();
	activeplayers.clear();
	rebirthwait.clear();
	Audio::stopSound( Audio::SOUND_GAME_OVER );

	// and shots
	au::Automaton::kill_shots();

	// and enemies
	killenemies = false;
	au::Automaton::kill_enemies();

	// and remaining elements
	au::Grid::finalize();
	au::Automaton::kill_remaining();

	// starting intervals
	count_gravity = rc::gravity_int();
	count_pursuit = rc::pursuit_int();
	count_dodging = rc::dodging_int();
	count_collide = rc::collide_int();
}
// (re)starts the game
void Main::start(){
	// cleaning
	clear();

	// restoring initial state
	au::Automaton::init();
	au::EnemyScheduler::init();
	au::Grid::init();

	// mercy state
	powerup_lifes = rc::powerup_lifes();

	// stats
	score = STARTING_SCORE;
	gameover = false;
	lifes = powerup_lifes ? STARTING_LIFES : 0;
	bombs = STARTING_BOMBS;
	gui::GUI::update();

	// playing sound and starting player
	Audio::playSound( Audio::SOUND_GAME_START );
	au::Automaton::create_player( 0 );
}

// computes the savegame file name
void Main::fill_sg(){
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
	sg = SG_NAME;
#else // Windows
	char *home = std::getenv( "HOME" );
	if ( home != NULL ){
		sg = home;
		sg += "/" SG_NAME;
	}else
		std::cerr << "ERROR: cannot get the name for the savegame file" << std::endl;
#endif // Not Windows
}
// serializes the game
void Main::serialize(){
	// no saves when game over or no file
	if ( gameover || sg.empty() )
		return;

	try{
		// opening file
		std::ofstream out( sg.c_str(), std::ofstream::out|std::ofstream::trunc );
		out.exceptions( std::ofstream::eofbit | std::ofstream::failbit | std::ofstream::badbit );

		// storing a savegame version
		out << SAVEGAME_VERSION << std::endl;

		// saving stats
		out << score << " " << lifes << " " << bombs << " " << powerup_lifes << std::endl;
		// saving rebirth wait
		out << rebirthwait.size();
		foreach( const std::pair<int_int>& r, rebirthwait )
			out << " " << r.first << " " << r.second;
		out << std::endl;
		// saving enemy scheduler
		au::EnemyScheduler::serialize( out );
		// saving automatons
		au::Automaton::sserialize( out );
	}catch( const std::ofstream::failure& err ){
		std::cerr << "ERROR: cannot save the game at file \"" << sg << "\"" << std::endl;
	}
}
// unserializes the game
void Main::unserialize(){
	// no load when no file
	if ( sg.empty() )
		return;

	// opening file
	std::ifstream in( sg.c_str(), std::ifstream::in );
	try{
		in.exceptions( std::ifstream::eofbit | std::ifstream::failbit | std::ifstream::badbit );

		// checking that the savegame version is known
		int sgv;
		in >> sgv;
		if ( sgv != SAVEGAME_VERSION ){
			std::cerr << "ERROR: saved game at file \"" << sg << "\" has an unknown version." << std::endl;
			return;
		}
	}catch( const std::ifstream::failure& err ){
		std::cerr << "ERROR: cannot load a saved game from file \"" << sg << "\"" << std::endl;
		return;
	}

	bool fine = false;
	try{
		// kill everything
		clear();

		// loading stats
		in >> score >> lifes >> bombs >> powerup_lifes;
		if ( (score < 0) || (score > MAX_SCORE) )
			throw UNSERIAL_SCORE;
		if ( (lifes < 0) || (lifes > MAX_LIFES) )
			throw UNSERIAL_LIFES;
		if ( (bombs < 0) || (bombs > MAX_BOMBS) )
			throw UNSERIAL_BOMBS;
		if ( (powerup_lifes < 0) || (!powerup_lifes && lifes) )
			throw UNSERIAL_POWERUP_LIFES;
		gameover = false;
		// loading rebirth wait
		int rws;
		in >> rws;
		if ( (rws < 0) || (rws > MAX_PLAYER_AMOUNT) )
			throw UNSERIAL_REBIRTHWAIT;
		for( ; rws ; --rws ){
			int time;
			int id;
			in >> time >> id;
			if ( time < 0 )
				throw UNSERIAL_REBIRTHWAIT_TIME;
			if ( (id < 0) || (id >= static_cast<int>(Input::get_inputs().size())) )
				throw UNSERIAL_REBIRTHWAIT_ID;
			rebirthwait.push_back( std::pair<int,int>(time, id) );
		}
		// loading enemy scheduler
		au::EnemyScheduler::unserialize( in );
		// loading automatons
		au::Automaton::sunserialize( in );

		fine = true;
		if ( score > hiscore )
			hiscore = score;
		gui::GUI::update();
	}catch( const std::ifstream::failure& err ){
		std::cerr << "ERROR: I/O error while reading saved game from file \"" << sg << "\"" << std::endl;
	}catch( const unserialize_error_t& err ){
		std::cerr << "ERROR: saved game at file \"" << sg << "\" is corrupted [" << err << "]" << std::endl;
	}

	// cannot load and we've lost present game, an empty one must be started
	if ( ! fine )
		start();
}

// program information
void Main::print_notice(){
	std::cout << PACKAGE_STRING << std::endl
	          << "Copyright (C) 2008 Carles Creus Lopez" << std::endl
	          << PACKAGE_NAME " is free software; you can redistribute it and/or modify it" << std::endl
	          << "under the terms of the GNU General Public License as published by" << std::endl
	          << "the Free Software Foundation; either version 3 of the License, or" << std::endl
	          << "(at your option) any later version." << std::endl
	          << "This program is distributed in the hope that it will be useful," << std::endl
	          << "but WITHOUT ANY WARRANTY; without even the implied warranty of" << std::endl
	          << "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << std::endl
	          << "GNU General Public License for more details." << std::endl
	          << "You should have received a copy of the GNU General Public License" << std::endl
	          << "along with this program. If not, see <http://www.gnu.org/licenses/>" << std::endl
	          << "or write to the Free Software Foundation, Inc., 51 Franklin Street," << std::endl
	          << "Fifth Floor, Boston, MA 02110-1301 USA." << std::endl;
}
// reading events since last round
void Main::read_events(){
	SDL_Event event;

	// starting an input round (clean the button information)
	Input::newround();

	while( SDL_PollEvent( &event ) ){
		switch(event.type){
			// check visiblity
			case SDL_ACTIVEEVENT:
				if ( event.active.state & SDL_APPACTIVE ){
					hidden = ! event.active.gain;
					if ( ! hidden )
						Timer::init();
				}
				break;
			// key processing
			case SDL_KEYDOWN:
				if ( event.key.keysym.sym != SDLK_PRINT ){
					Input::key( event.key.keysym.sym, true );
					gui::Menu::captureKey( event.key.keysym.sym );
				}else
					screenshot = true; // screenshot key is hardcoded, but it is way more obvious this way
				break;
			case SDL_KEYUP:
				if ( event.key.keysym.sym != SDLK_PRINT )
					Input::key( event.key.keysym.sym, false );
				break;
			// joystick axis processing
			case SDL_JOYAXISMOTION:
				Input::axis( event.jaxis.which, event.jaxis.axis, event.jaxis.value );
				gui::Menu::captureAxis( event.jaxis.which, event.jaxis.axis, event.jaxis.value );
				break;
			// joystick button processing
			case SDL_JOYBUTTONDOWN:
				Input::button( event.jbutton.which, event.jbutton.button, true );
				gui::Menu::captureButton( event.jbutton.which, event.jbutton.button );
				break;
			case SDL_JOYBUTTONUP:
				Input::button( event.jbutton.which, event.jbutton.button, false );
				break;
			// player is exiting, although not from the menu
			case SDL_QUIT:
				quit( NORMAL_EXIT );
				break;
		}
	}
}
// draws everything (from back to front)
void Main::draw(){
#ifndef NDEBUG
	// partition information
	au::Automaton::draw_map();
#endif
	// drawing grid
	au::Grid::draw();
	// drawing objects
	au::Automaton::draw();
	// drawing gui
	gui::GUI::draw();
	gui::Menu::draw();
}
// processing intervals
int Main::count_gravity;
int Main::count_pursuit;
int Main::count_dodging;
int Main::count_collide;
// process everything
void Main::process( int milliseconds ){
	// cleaning enemies
	if ( killenemies ){
		killenemies = false;
		au::Automaton::kill_enemies();
	}

	// menu == pause
	if ( ! gui::Menu::is_on() ){
		// spawning more enemies
		if ( au::Automaton::players_amount() > 0 )
			au::EnemyScheduler::update( milliseconds );

		// updating gravities
		if ( timetrigger( rc::gravity_int(), milliseconds, count_gravity ) )
			au::Automaton::gravitate();

		// updating pursuiters
		if ( timetrigger( rc::pursuit_int(), milliseconds, count_pursuit ) )
			au::Automaton::pursuit();

		// updating dodgers
		if ( timetrigger( rc::dodging_int(), milliseconds, count_dodging ) )
			au::Automaton::dodge();

		// always processing everything
		au::Automaton::process( milliseconds );
		// always killing zombies
		au::Automaton::kill_zombies();

		// always trying to create players
		entering_scheduler();
		rebirth_scheduler( milliseconds );

		// checking collisions
		if ( timetrigger( rc::collide_int(), milliseconds, count_collide ) )
			au::Automaton::collide();
	}

	// menu
	gui::Menu::process( milliseconds );
}
// checks if the interval has already elapsed
bool Main::timetrigger( int interval, int milliseconds, int& count ){
	if ( interval ){
		// if not in real time
		if ( interval < 0 ){
			// if it counts processing iterations
			++count;
			if ( count == 0 ){
				// it has reached its goal
				count = interval;
				return true;
			}else
				return false;
		}else{
			// else it counts milliseconds
			count -= milliseconds;
			if ( count <= 0 ){
				count = interval;
				return true;
			}else
				return false;
		}
	}else{
		// if in real time, compute always
		return true;
	}
}
// active player's id
std::set<int> Main::activeplayers;
// scheduled players
std::list< std::pair<int,int> > Main::rebirthwait;
// player creation when presses bomb button
void Main::entering_scheduler(){
	// if lifes left and someone pressed bomb
	if ( lifes && Input::get_agg_bomb() ){
		int p = 0;
		// for each input
		foreach( const Input *i, Input::get_inputs() ){
			// if an inactive player has pressed bomb, create player
			if ( i->get_bomb() && !activeplayers.count( p ) ){
				au::Automaton::create_player( p );
				--lifes;
				gui::GUI::updateLifes();

				if ( ! lifes )
					break;
			}
			++p;
		}
	}
}
// player creation when killed
void Main::rebirth_scheduler( int milliseconds ){
	if ( rebirthwait.empty() )
		return;

	// counting down time for each scheduled rebirth
	foreach( std::pair<int_int>& r, rebirthwait ){
		r.first -= milliseconds;

		// creating player if count has ended
		if ( r.first < 0 )
			au::Automaton::create_player( r.second );
	}

	// cleaning list
	while( !rebirthwait.empty() && (rebirthwait.front().first < 0) )
		rebirthwait.pop_front();
}
// updates player information
void Main::add_player( int id ){
	activeplayers.insert( id );
}
void Main::remove_player( int id ){
	if ( lifes ){
		// if we have some lifes left, rebirth
		--lifes;
		gui::GUI::updateLifes();

		rebirthwait.push_back( std::pair<int,int>( rc::player_reb_t(), id ) );
	}else{
		// one active player less
		activeplayers.erase( id );

		if ( activeplayers.empty() ){
			gameover = true;
			gui::GUI::updateGameover();
			Audio::stopSound( Audio::SOUND_PLAYER_DEATH );
			Audio::playSound( Audio::SOUND_GAME_OVER );
		}
	}

	// if no players alive, we do not want enemies
	if ( rebirthwait.size() == activeplayers.size() ){
		killenemies = true;
		au::EnemyScheduler::stop();
	}
}
// throws a bomb at xy
void Main::bombing( const util::Vector& xy ){
	if ( bombs ){
		--bombs;
		gui::GUI::updateBombs();

		killenemies = true;

		Audio::playSound( Audio::SOUND_BOMB );

		au::Automaton::create_blast( xy );
	}
}
// computes game activity: 1.0=normal
real Main::activity(){
	real playermult;
	real  scoremult;
	real   shotmult;

	// player amount multiplier
	playermult = 1.0 + (au::Automaton::players_amount()-1)*rc::player_increment();

	// score multiplier
	scoremult = 1.0 + score*rc::score_revamping();

	// shot/enemy ratio reduction
	if ( !au::Automaton::shots_amount() || !au::Automaton::enemies_amount() )
		shotmult = 1.0;
	else{
		shotmult = std::min( static_cast<real>(1.0),
		                     (rc::game_overwhelming()
		                      * static_cast<real>(au::Automaton::shots_amount()))
		                     / static_cast<real>(au::Automaton::enemies_amount()) );
	}

	return std::min( rc::max_activity(), std::max( rc::min_activity(), playermult*scoremult*shotmult ) );
}
// checks the arguments
void Main::check_args( std::list<std::string> args ){
	if ( args.empty() )
		return;

	bool do_quit = false;

	while( ! args.empty() ){
		std::string arg = args.front();

		// version flag
		if ( (arg == "-v" ) || (arg == "--version") ){
			print_notice();
			do_quit = true;
		// help flag
		}else if ( (arg == "-h" ) || (arg == "-?") || (arg == "--help") ){
			std::cout << "Possible arguments are:" << std::endl
			          << "\011-h, -?, --help" << std::endl
			          << "\011-v, --version" << std::endl
			          << "\011-Dx[=y]" << std::endl
			          << "\011--slow" << std::endl
			          << "\011--fast" << std::endl
			          << "\011--easy" << std::endl
			          << "\011--hard" << std::endl
			          << "\011--merciless" << std::endl
			          << "Any other argument is incorrect." << std::endl
			          << std::endl
			          << "Check 'man " PACKAGE_TARNAME "' for gameplay information." << std::endl;
			do_quit = true;
		// a command-line define
		}else if ( (arg.size() >= 3) && (arg.substr(0,2) == "-D") ){
			rc::add_define( arg.substr(2) );
		// standard modifiers
		}else if ( (arg == "--slow") || (arg == "--fast") || (arg == "--easy") || (arg == "--hard") || (arg == "--merciless") ){
			rc::add_define( arg.substr(2) );
		// unknown flag
		}else{
			std::cerr << arg << ": unknown argument, see --help" << std::endl;
			do_quit = true;
		}

		// done with present flag
		args.remove( arg );
	}

	if ( do_quit )
		quit( NORMAL_EXIT );
}

// program start
void Main::main( int argc, char *argv[] ){
	// checking args
	check_args( std::list<std::string>( argv+1, argv+argc ) );

	// printing notice
	print_notice();
	std::cout << std::endl;

	// starting all needed modules
	if ( SDL_Init( SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO | SDL_INIT_JOYSTICK ) < 0 ){
		std::cerr << "FATAL: cannot initialize SDL library: " << SDL_GetError() << std::endl;
		quit( ERROR_INIT );
	}

	// stopping key repeat
	if ( SDL_EnableKeyRepeat( 0, 0 ) == -1 ){
		std::cerr << "FATAL: cannot ignore key repetitions: " << SDL_GetError() << std::endl;
		quit( ERROR_KREPEAT );
	}
	// ignoring useless devices
	SDL_EventState( SDL_MOUSEMOTION, SDL_IGNORE );
	SDL_EventState( SDL_MOUSEBUTTONDOWN, SDL_IGNORE );
	SDL_EventState( SDL_MOUSEBUTTONUP, SDL_IGNORE );
	SDL_EventState( SDL_VIDEORESIZE, SDL_IGNORE );
	SDL_EventState( SDL_VIDEOEXPOSE, SDL_IGNORE );

	// starting modules in proper order (see their dependencies)
	start_modules();

	hiscore = Config::get_hiscore(); // need to get it just once, the first time

	// getting savegame file name
	fill_sg();

	// playing
	start();
	for(;;){
		screenshot = false;

		read_events();

		if ( ! hidden ){
			process( Timer::incr() );

			vis::Screen::start_drawing();
			draw();
			if ( screenshot ){
				vis::Screen::takescreenshot();
				// a screenshot may take too long, so reseting timer
				Timer::init();
			}
			vis::Screen::end_drawing();
		}
	}
}

// SDL program start
int main( int argc, char *argv[] ){
	Main::main( argc, argv );
	return -1;
}
