// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cmath>
#include <cstdlib>
#include <iostream>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#define int_Automatonp int,Automaton*
#include "../rc/Constants.hpp"
#include "../util/Random.hpp"
#include "../vis/Draw.hpp"
#include "../Audio.hpp"
#include "../Config.hpp"
#include "../Main.hpp"
#include "Grid.hpp"
#include "Automaton.hpp"

// ctor
au::Automaton::Automaton( const util::Vector& xy, const util::Vector& v, real r, int t ) :
 xy( xy ),
 v( v ),
 r( r ),
 t( t ),
 m( 0.0 ),
 g(),
 dolitter( true ),
 ringdie( false ),
 trackable( NO_TRACKABLE ),
 tracking_xy(),
 tracking_r(),
 visiblityorder( NO_VISIBLE ),
 pursuiter( false ),
 processable( false ),
 gravityprovider( false ),
 gravitylistener( false ),
 dodger( false ),
 collided_types( NO_COLLIDER ),
 collideractive_type( NO_COLLIDER ),
 enemy( false ),
 player( false ),
 shot( false )
{
}
// dtor
au::Automaton::~Automaton(){
	if ( visiblityorder )
		Automaton::visibles.erase( std::pair<int,Automaton*>( visiblityorder, this ) );
	if ( pursuiter )
		Automaton::pursuiters.erase( this );
	if ( processable ){
		Automaton::deferredproc.erase( this );
		Automaton::processables.erase( this );
	}
	if ( gravityprovider )
		Automaton::gravityproviders.erase( this );
	if ( gravitylistener )
		Automaton::gravitylisteners.erase( this );
	if ( dodger )
		Automaton::dodgers.erase( this );
	if ( collided_types )
		Automaton::colliderspassive->erase( this, collided_types, xy, r );
	if ( collideractive_type )
		Automaton::collidersactive.erase( this );
	if ( enemy )
		Automaton::enemies.erase( this );
	if ( player )
		Automaton::players.erase( this );
	if ( shot )
		Automaton::shots.erase( this );
}
// initializes the automaton [depends on: Constants]
void au::Automaton::init(){
	if ( colliderspassive == NULL )
		colliderspassive = new util::Partition<Automaton*>( rc::colliderpassive_resolution(), COL_AMOUNT );

	shot_counter = 0;
	shot_points[NORMAL_SHOTS]    = rc::normal_shots_int();
	shot_points[FAST_SHOTS]      = rc::fast_shots_int() + shot_points[NORMAL_SHOTS];
	shot_points[MULTI_SHOTS]     = rc::multi_shots_int() + shot_points[FAST_SHOTS];
	shot_points[LASER_SHOTS]     = rc::laser_shots_int() + shot_points[MULTI_SHOTS];
	shot_points[LIGHTNING_SHOTS] = rc::lightning_shots_int() + shot_points[LASER_SHOTS];
	max_count = shot_points[SHOTS_AMOUNT-1];
	shot_rates[NORMAL_SHOTS]    = rc::normal_shots_rate();
	shot_rates[FAST_SHOTS]      = rc::fast_shots_rate();
	shot_rates[MULTI_SHOTS]     = rc::multi_shots_rate();
	shot_rates[LASER_SHOTS]     = rc::laser_shots_rate();
	shot_rates[LIGHTNING_SHOTS] = rc::lightning_shots_rate();
	next_shottype( 0 );
}
// ends the Automaton
void au::Automaton::finalize(){
	kill_players();
	kill_shots();
	kill_enemies();

	if ( colliderspassive != NULL )
		delete colliderspassive;
}
// serializes the automatons
void au::Automaton::sserialize( std::ostream& out ){
	// saving static parameters
	out << shot_counter;
	for( int i=0 ; i<SHOTS_AMOUNT ; ++i )
		out << " " << shot_points[i];
	out << std::endl;

	// saving dynamic objects
	au::Grid::serialize( out );
	out << processables.size() << std::endl;
	foreach( Automaton *p, processables )
		p->serialize( out );
	au::Grid::end_serialize();
}
// unserializes the automatons
void au::Automaton::sunserialize( std::istream& in ){
	// loading static parameters
	in >> shot_counter;
	for( int i=0 ; i<SHOTS_AMOUNT ; ++i ){
		in >> shot_points[i];
		if ( i ){
			if ( shot_points[i] < shot_points[i-1] )
				throw Main::UNSERIAL_SHOT_POINTS;
		}else{
			if ( shot_points[i] < 0 )
				throw Main::UNSERIAL_SHOT_POINTS;
		}
	}
	max_count = shot_points[SHOTS_AMOUNT-1];
	if ( max_count == 0 )
		throw Main::UNSERIAL_SHOT_POINTS;
	if ( (shot_counter < 0) || (shot_counter >= max_count) )
		throw Main::UNSERIAL_SHOT_COUNTER;
	next_shottype( 0 );

	// loading dynamic objects
	au::Grid::unserialize( in );
	int ps;
	in >> ps;
	if ( ps < 0 )
		throw Main::UNSERIAL_PROCESSABLES;
	for( ; ps ; --ps ){
		int id;
		in >> id;
		switch( id ){
			case BLAST_SERIAL:
				Blast::unserialize( in );
				break;
			case BULLET_SERIAL:
				Bullet::unserialize( in );
				break;
			case DEBRIS_SERIAL:
				Debris::unserialize( in );
				break;
			case EXAMINIS_SERIAL:
				Examinis::unserialize( in );
				break;
			case LAGVARD_SERIAL:
				Lagvard::unserialize( in );
				break;
			case LASER_SERIAL:
				Laser::unserialize( in );
				break;
			case LASERTAIL_SERIAL:
				LaserTail::unserialize( in );
				break;
			case LIGHTNING_SERIAL:
				Lightning::unserialize( in );
				break;
			case PLAYER_SERIAL:
				Player::unserialize( in );
				break;
			case PROTON_SERIAL:
				Proton::unserialize( in );
				break;
			case REDHOLE_SERIAL:
				RedHole::unserialize( in );
				break;
			case SCARED_SERIAL:
				Scared::unserialize( in );
				break;
			case SHIELDED_SERIAL:
				Shielded::unserialize( in );
				break;
			case SPAWN_SERIAL:
				Spawn::unserialize( in );
				break;
			case SPAWNSHIP_SERIAL:
				Spawnship::unserialize( in );
				break;
			default:
				throw Main::UNSERIAL_ID;
		}
	}
}

// starts the visible object
void au::Automaton::start_visible( int visiblityorder ){
	this->visiblityorder = visiblityorder;
	Automaton::visibles.insert( std::pair<int,Automaton*>( visiblityorder, this ) );
}
// visible automatons
std::set< std::pair<int,au::Automaton*> > au::Automaton::visibles;
// drawing function
void au::Automaton::draw(){
	foreach( const std::pair<int_Automatonp>& v, visibles ){
		vis::Draw::reset();
		v.second->tracking_save();
		v.second->specific_draw();
		v.second->tracking_check();
	}
}
#ifndef NDEBUG
void au::Automaton::draw_map(){
	vis::Draw::reset();
	colliderspassive->draw();
}
#endif
// starts the pursuiter object
void au::Automaton::start_pursuiter(){
	pursuiter = true;
	Automaton::pursuiters.insert( this );
}
// player following automatons
std::set<au::Automaton*> au::Automaton::pursuiters;
// updates pursuiters
void au::Automaton::pursuit(){
	// for each follower
	foreach( Automaton *p, pursuiters ){
		real nearest2 = REAL_MAX;
		util::Vector itis;

		// search closest player
		foreach( Automaton *pl, players ){
			util::Vector dist = p->distance_to( pl );
			real mag2 = dist.magnitude2();

			if ( (mag2 > 0.0) && (mag2 < nearest2) ){
				nearest2 = mag2;
				itis = dist;
			}
		}

		real nearest = std::sqrt( nearest2 );

		p->tracking_save();
		p->specific_there( itis/nearest, nearest );
		p->tracking_check();
	}
}
// starts the processable object
void au::Automaton::start_processable(){
	processable = true;
	if ( Automaton::deferproc )
		Automaton::deferredproc.insert( this );
	else
		Automaton::processables.insert( this );
}
// ends the processable object
void au::Automaton::end_processable(){
	processable = false;
	Automaton::deferredproc.erase( this );
	Automaton::processables.erase( this );
}
// while processing objects, processable objects may be created
// so it needs to be deferred to avoid corrupting the iteration
bool au::Automaton::deferproc = false;
std::set<au::Automaton*> au::Automaton::deferredproc;
// processable automatons
std::set<au::Automaton*> au::Automaton::processables;
// process everything
void au::Automaton::process( int milliseconds ){
	deferproc = true;
	foreach( Automaton *p, processables ){
		p->tracking_save();
		p->t += milliseconds;
		p->specific_process( milliseconds );
		p->tracking_check();
	}
	processables.insert( deferredproc.begin(), deferredproc.end() );
	deferredproc.clear();
	deferproc = false;
}
// starts the gravity provider object
void au::Automaton::start_gravityprovider(){
	gravityprovider = true;
	Automaton::gravityproviders.insert( this );
}
// starts the gravity listener object
void au::Automaton::start_gravitylistener(){
	gravitylistener = true;
	Automaton::gravitylisteners.insert( this );
}
// gravity listener automatons
std::set<au::Automaton*> au::Automaton::gravitylisteners;
// gravity provider automatons
std::set<au::Automaton*> au::Automaton::gravityproviders;
// compute gravitational stuff
void au::Automaton::gravitate(){
	// for each listener
	foreach( Automaton *l, gravitylisteners ){
		// reset its field
		l->g = util::Vector(0.0, 0.0);

		// for each provider
		foreach( Automaton *p, gravityproviders ){
			if ( p == l )
				continue;

			// update its field
			util::Vector w = l->distance_to(p); // vector from l to p
			real d = w.magnitude(); // distance magnitude
			if ( ! std::isnormal(d) )
				continue; // ignoring weird values
			real gm = std::min( rc::max_g(), rc::constant_g() * p->m / d ); // g magnitude
			l->g += w.magnitude(gm);
		}
	}
}
// starts the dodger object
void au::Automaton::start_dodger(){
	dodger = true;
	Automaton::dodgers.insert( this );
}
// shot dodgers
std::set<au::Automaton*> au::Automaton::dodgers;
// dodge shots
void au::Automaton::dodge(){
	// for each dodger
	foreach( Automaton *d, dodgers ){
		util::Vector f(0.0, 0.0);
		real fr, fi;
		d->specific_dodger_parameters( fr, fi );

		// for each shot
		foreach( Automaton *s, shots ){
			if ( s == d )
				continue;

			// update fear
			util::Vector w = s->distance_to(d); // vector from s to d
			real m = w.magnitude(); // distance magnitude
			if ( !std::isnormal(m) || (m > fr) )
				continue; // it is not considered dangerous
			real fm = std::min( rc::max_fear(), fi / m );
			f += w.magnitude(fm);
		}

		d->tracking_save();
		d->specific_dodger( f );
		d->tracking_check();
	}
}
// starts the passive collider object
void au::Automaton::start_passivecollider( int collided_types ){
	this->collided_types = collided_types;
	Automaton::colliderspassive->insert( this, collided_types, xy, r );
	tracking_start( TRACKABLE_COLLIDERPASSIVE );
}
// starts the active collider object
void au::Automaton::start_activecollider( int collideractive_type ){
	this->collideractive_type = collideractive_type;
	Automaton::collidersactive.insert( this );
}
// ends the active collider object
void au::Automaton::end_activecollider(){
	collideractive_type = NO_COLLIDER;
	Automaton::collidersactive.erase( this );
}
// active collider automatons
std::set<au::Automaton*> au::Automaton::collidersactive;
// passive colliders automatons
util::Partition<au::Automaton*> *au::Automaton::colliderspassive = NULL;
// compute collisions
void au::Automaton::collide(){
	// if there are no automatons that may collide, nothing to do
	if ( collidersactive.empty() || colliderspassive->empty() )
		return;

	// saving the automatons that die
	std::set<Automaton*> dead; // unordered

	// for each active collider
	foreach( Automaton *a, collidersactive ){
		if ( dead.find(a) != dead.end() )
			continue;

		// getting collision candidates
		std::list<Automaton*> cands;
		colliderspassive->candidates( a->collideractive_type, a->xy, a->r, cands );

		a->tracking_save();
		// check if it collides with any of its passive colliders
		foreach( Automaton *p, cands ){
			if ( p == a )
				continue;
			if ( dead.find(p) != dead.end() )
				continue;
			if ( ! p->intersects_automaton(a) )
				continue;

			p->tracking_save();
			int ans = NO_DEATHS;
			p->send_collision( ans, a );
			p->tracking_check();
			// killing who has died
			switch(ans){
				case ACTIVE_DIES:
					dead.insert(a);
					break;
				case PASSIVE_DIES:
					dead.insert(p);
					break;
				case BOTH_DIE:
					dead.insert(a);
					dead.insert(p);
					break;
			}
			if ( ans & ACTIVE_DIES )
				break;
		}
		a->tracking_check();
	}

	// removing dead automatons
	foreach( Automaton *a, dead )
		delete a;
}
// creates a spring object at the given location
au::Spring *au::Automaton::create_spring( real x, real y ){
	return new Spring( util::Vector(x,y) );
}
// creates a blast at the given location
void au::Automaton::create_blast( const util::Vector& xy ){
	new Blast( xy );
}
// litters the screen with some random-color debris, at a radius distance from xy in
// omega direction (which has an imprecision of 2phi).
// The amount of debris is max(1, ceil(3 2phi)), i.e. 3 per radiant, at least 1
void au::Automaton::create_debris( const util::Vector& xy, real radius, real omega, real phi ){
	create_debris( xy, radius, omega, phi, true, 0.0, 0.0, 0.0 );
}
// litters the screen with a circle of debris of the specified color, at a radius
// distance from xy
// The amount of debris is max(1, ceil(3 2phi)), i.e. 3 per radiant, at least 1
void au::Automaton::create_debris( const util::Vector& xy, real radius, real red, real green, real blue ){
	create_debris( xy, radius, 0.0, PI, false, red, green, blue );
}
// litters the screen with some debris, at a radius distance from xy in omega direction
// (which has an imprecision of 2phi).
// The amount of debris is max(1, ceil(3 2phi)), i.e. 3 per radiant, at least 1
void au::Automaton::create_debris( const util::Vector& xy, real radius, real omega, real phi,
                                   bool randcolors, real red, real green, real blue ){
	if ( Config::get_debris() ){
		// for all the debris objects that must be created
		for( int i = std::max( 1, static_cast<int>( std::ceil(2.0*phi*rc::debris_per_radiant()) ) );
		     i;
		     --i
		   ){
			// computing its color if it is random
			if ( randcolors ){
				red   = util::Random::get( rc::debris_min_red(),   rc::debris_max_red() );
				green = util::Random::get( rc::debris_min_green(), rc::debris_max_green() );
				blue  = util::Random::get( rc::debris_min_blue(),  rc::debris_max_blue() );
			}

			// getting a random direction
			util::Vector dir = util::Vector(1.0).angle(omega + util::Random::get(phi));

			new Debris( xy+(dir*radius), dir*rc::debris_s(), red, green, blue );
		}
	}
}

// starts the enemy object
void au::Automaton::start_enemy( bool play ){
	enemy = true;
	Automaton::enemies.insert( this );
	if ( play )
		Audio::playSound( Audio::SOUND_ENEMY_BIRTH );
}
// enemy automatons
std::set<au::Automaton*> au::Automaton::enemies;
// creates an enemy identified by code at location xy
void au::Automaton::create_enemy( int code, const util::Vector& xy ){
	switch(code){
		case EXAMINIS_CODE:
			new Examinis( xy );
			break;
		case LAGVARD_CODE:
			new Lagvard( xy );
			break;
		case REDHOLE_CODE:
			new RedHole( xy );
			break;
		case SCARED_CODE:
			new Scared( xy );
			break;
		case SHIELDED_CODE:
			new Shielded( xy );
			break;
		case SPAWNSHIP_CODE:
			new Spawnship( xy );
			break;
	}
}
// creates protons spawned from inside the redhole when it dies
void au::Automaton::create_protons( const util::Vector& xy, real r, real m ){
	r = std::max( r, rc::redhole_r()*rc::redhole_birth_scale() );
	for( int i = static_cast<int>( std::ceil(rc::proton_proportion() * ((m-rc::redhole_mass()-rc::redhole_mass_t()) / rc::redhole_mass())) );
	     i;
	     --i
	   ){
		// getting a random location inside the redhole
		util::Vector p_xy = xy + util::Vector( util::Random::get(r) ).angle( util::Random::get(PI) );
		if ( player_free_area( p_xy, rc::small_spawning_area_r() ) )
			new Proton( p_xy );
	}
}
// creates the spawnship's three spawned children
void au::Automaton::create_spawned( const util::Vector& xy ){
	Spawn *s = new Spawn( xy );
	s->add_child( new Spawned( s, xy, 0.0*(2.0*PI/3.0) ) );
	s->add_child( new Spawned( s, xy, 1.0*(2.0*PI/3.0) ) );
	s->add_child( new Spawned( s, xy, 2.0*(2.0*PI/3.0) ) );
}
// creates the examinis' apiculas
au::ExaminisApicula *au::Automaton::create_examinisapicula( Examinis *swarmer, const util::Vector& xy, bool queen ){
	return new ExaminisApicula( swarmer, xy, queen );
}

// starts the player object
void au::Automaton::start_player(){
	player = true;
	Automaton::players.insert( this );
}
// alive players
std::set<au::Automaton*> au::Automaton::players;
// creates a player controled by user id
void au::Automaton::create_player( int id ){
	new Player( id );
}

// starts the shot object
void au::Automaton::start_shot(){
	shot = true;
	Automaton::shots.insert( this );
	dolitter = false;
}
// shots
std::set<au::Automaton*> au::Automaton::shots;
// present type for shots
int au::Automaton::shot_type;
// counts the number of score points acumulated
int au::Automaton::shot_counter;
// score point intervals between each type of shot
int au::Automaton::shot_points[SHOTS_AMOUNT];
// modulus of the counter
int au::Automaton::max_count;
// shot rates for each type of shot
int au::Automaton::shot_rates[SHOTS_AMOUNT];
// sets the shots to the starting value
void au::Automaton::reset_shottype(){
	shot_counter = 0;
	next_shottype( 0 );
}
// updates the shot type
void au::Automaton::next_shottype( int score_incr ){
	shot_counter = (shot_counter + score_incr) % max_count;

	if ( shot_counter < shot_points[NORMAL_SHOTS] )
		shot_type = NORMAL_SHOTS;
	else if ( shot_counter < shot_points[FAST_SHOTS] )
		shot_type = FAST_SHOTS;
	else if ( shot_counter < shot_points[MULTI_SHOTS] )
		shot_type = MULTI_SHOTS;
	else if ( shot_counter < shot_points[LASER_SHOTS] )
		shot_type = LASER_SHOTS;
	else if ( shot_counter < shot_points[LIGHTNING_SHOTS] )
		shot_type = LIGHTNING_SHOTS;
}
// gets the present milliseconds lapse between consecutive shots
int au::Automaton::shot_rate(){
	return shot_rates[shot_type];
}
// creates the present shot type at a radius distance from the origin in the given direction
void au::Automaton::create_shot( const util::Vector& origin, const util::Vector& direction, real radius ){
	real aux_angle_o;
	real aux_angle_i;

	switch(shot_type){
		case NORMAL_SHOTS:
			new Bullet( origin+(direction*radius), direction*rc::bullet_s() );
			break;
		case FAST_SHOTS:
			new Bullet( origin+(direction*radius), direction*rc::fast_bullet_s() );
			break;
		case MULTI_SHOTS:
			aux_angle_o = direction.angle() - rc::multi_shots_range()/2.0;
			aux_angle_i = rc::multi_shots_range()/(rc::multi_shots_amount()-1);
			for( int i=0 ; i<rc::multi_shots_amount() ; ++i ){
				util::Vector aux_direction = util::Vector(1.0).angle( aux_angle_o + i*aux_angle_i );
				new Bullet( origin+(aux_direction*radius), aux_direction*rc::multi_bullet_s() );
			}
			break;
		case LASER_SHOTS:
			new Laser( origin+(direction*radius), direction*rc::laser_s(), direction*rc::laser_tail_s() );
			break;
		case LIGHTNING_SHOTS:
			new Lightning( origin+(direction*radius), direction*rc::lightning_s() );
			break;
	}
}
// creates a tail for the laser
void au::Automaton::create_lasertail( const util::Vector& xy, const util::Vector& v2, LaserParts *parts ){
	new LaserTail( xy, v2, parts );
}
// creates a lightning spark
au::LightningSpark *au::Automaton::create_lightningspark( Lightning *swarmer, const util::Vector& xy,
                                                          const util::Vector& v, int lifes, int nttl, int turn ){
	return new LightningSpark( swarmer, xy, v, lifes, rc::lightning_t()-nttl, turn );
}

// called when an automaton wants to commit suicided
void au::Automaton::add_zombie(){
	Automaton::zombies.insert( this );
}
// dead automatons
std::set<au::Automaton*> au::Automaton::zombies;

// automaton killers
void au::Automaton::kill_enemies(){
	while( ! enemies.empty() )
		delete *(enemies.begin());
}
void au::Automaton::kill_players(){
	while( ! players.empty() )
		delete *(players.begin());
}
void au::Automaton::kill_shots(){
	while( ! shots.empty() )
		delete *(shots.begin());
	Audio::stopSound( Audio::SOUND_BULLET_HIT );
}
void au::Automaton::kill_zombies(){
	foreach( Automaton *a, zombies )
		delete a;
	zombies.clear();
}
void au::Automaton::kill_remaining(){ // callable when only blasts, debris, lasertails and lightnings remain
	while( ! processables.empty() )
		delete *(processables.begin());
}

// automaton amounts
int au::Automaton::enemies_amount(){
	return static_cast<int>( enemies.size() );
}
int au::Automaton::players_amount(){
	return static_cast<int>( players.size() );
}
int au::Automaton::shots_amount(){
	return static_cast<int>( shots.size() );
}

// checks if at xy with radius r there are no players
bool au::Automaton::player_free_area( const util::Vector& xy, real r ){
	foreach( const Automaton *a, players ){
		if ( a->intersects_circle( xy, r ) )
			return false;
	}
	return true;
}

// checks if this automaton intersects another automaton
bool au::Automaton::intersects_automaton( const Automaton *a ) const{
	return intersects_circle( a->xy, a->r );
}
// checks if this automaton intersects with the circle
bool au::Automaton::intersects_circle( const util::Vector& center, real radius ) const{
	return (xy-center).magnitude() < r+radius;
}
// intersection magnitude of the two automatons
real au::Automaton::intersection_mag( const Automaton *a ) const{
	return r + a->r - distance_to(a).magnitude();
}

// return the vector of the distance from this automaton to a
util::Vector au::Automaton::distance_to( const Automaton *a ) const{
	return a->xy - xy;
}

// adds this type of trackability to the automaton, to properly qualify update calls
void au::Automaton::tracking_start( int type ){
	trackable |= type;
}
// saves the information that needs to be tracked
void au::Automaton::tracking_save(){
	if ( trackable ){
		tracking_xy = xy;
		tracking_r  = r;
	}
}
// checks and notifies a change in the tracked data
void au::Automaton::tracking_check(){
	if ( trackable && ((tracking_r != r) || (tracking_xy != xy)) ){
		if ( trackable & TRACKABLE_COLLIDERPASSIVE )
			colliderspassive->update( this, collided_types, tracking_xy, tracking_r, xy, r );
	}
}

// sets that it must not litter when dying
void au::Automaton::no_litter(){
	dolitter = false;
}

// sets the position
void au::Automaton::set_xy( const util::Vector& newxy ){
	tracking_save();
	xy = newxy;
	if ( ringdie )
		checkRing_justdie();
	else
		checkRing_stop();
	tracking_check();
}
// gets current position
const util::Vector& au::Automaton::get_xy() const{
	return xy;
}
// gets mass
real au::Automaton::get_mass() const{
	return m;
}

// helper function for processing code that sets the automaton inside the ring
bool au::Automaton::checkRing_stop(){
	return checkRing( true, false, false, 0.0 );
}
// helper function for processing code that sets the automaton inside the ring
// but kills it if it was out
bool au::Automaton::checkRing_die(){
	return checkRing( true, true, false, 0.0 );
}
// helper function for processing code that kills the automaton if it was out
// of the ring
bool au::Automaton::checkRing_justdie(){
	return checkRing( false, true, false, 0.0 );
}
// helper function for processing code that sets the automaton inside the ring
// and makes it bounce if it has hit the ring limits
bool au::Automaton::checkRing_bounce( real nolosesratio ){
	return checkRing( true, false, true, nolosesratio );
}
// helper function for the processing code that checks if the automaton is
// still in the ring and performs some actions if not:
//   if move, sets it inside the ring changing xy
//   if die, it kills the automaton
//   if bounce, v direction is changed and multiplied by nolosesratio
bool au::Automaton::checkRing( bool move, bool die, bool bounce, real nolosesratio ){
	bool res = false;

	// checking x component
	real& x = xy.cx();
	real& vx = v.cx();
	if ( x < rc::ring_x0()+r ){
		// if too left
		res = true;
		if ( move )
			x = rc::ring_x0()+r;
		if ( bounce && (vx < 0.0) )
			vx = -vx * nolosesratio;
	}else if ( x > rc::ring_x1()-r ){
		// if too right
		res = true;
		if ( move )
			x = rc::ring_x1()-r;
		if ( bounce && (vx > 0.0) )
			vx = -vx * nolosesratio;
	}

	// checking y component
	real& y = xy.cy();
	real& vy = v.cy();
	if ( y < rc::ring_y0()+r ){
		// if too low
		res = true;
		if ( move )
			y = rc::ring_y0()+r;
		if ( bounce && (vy < 0.0) )
			vy = -vy * nolosesratio;
	}else if ( y > rc::ring_y1()-r ){
		// if too high
		res = true;
		if ( move )
			y = rc::ring_y1()-r;
		if ( bounce && (vy > 0.0) )
			vy = -vy * nolosesratio;
	}

	// if we had to die on a ring collision
	if ( res && die )
		add_zombie();

	return res;
}

// sets the final speeds of this automaton and automaton a when they ellastically
// bounce and their masses are m1 and m2 respectively
void au::Automaton::ellastic_collision( Automaton *a, real m1, real m2, real nolosesratio ){
	// unit normal vector
	util::Vector un = (a->xy - xy).magnitude(1.0);
	// unit tangent vector
	util::Vector ut = util::Vector(-un.cy(), un.cx());

	// v1 normal and tangent component
	real v1n = v * un;
	real v1t = v * ut;
	// v2 normal and tangent component
	real v2n = a->v * un;
	real v2t = a->v * ut;

	// prime normal components
	real v1n_p;
	real v2n_p;
	if ( m1 != m2 ){
		// when masses are different
		v1n_p = (v1n*(m1-m2) + 2.0*m2*v2n) / (m1+m2);
		v2n_p = (v2n*(m2-m1) + 2.0*m1*v1n) / (m1+m2);
	}else{
		// when masses are equal
		v1n_p = v2n;
		v2n_p = v1n;
	}

	// setting final speeds
	v    = un*(v1n_p*nolosesratio) + ut*v1t;
	a->v = un*(v2n_p*nolosesratio) + ut*v2t;
}
// moves the automatons to avoid the collision
void au::Automaton::avoid_collision( Automaton *a ){
	util::Vector movexy = distance_to(a).magnitude(intersection_mag(a)) * rc::collision_guiltyness();
	xy    -= movexy;
	a->xy += movexy;
}
// checks if a reflect of the automaton a is done and computes its resulting speed.
// with pass_range is specified the angle that does not reflect, ~0.0 for perpendicular
// and negative will always reflect
bool au::Automaton::reflect_collision( const Automaton *a, real pass_range, util::Vector& v2f ) const{
	// unit normal vector
	util::Vector un = (a->xy - xy).magnitude(1.0);
	// unit tangent vector
	util::Vector ut = util::Vector(-un.cy(), un.cx());

	// collision angle
	real theta = a->v.angle() - un.angle();
	if ( theta > PI )
		theta = std::abs(-2.0*PI + theta);
	else if ( theta < -PI )
		theta = std::abs( 2.0*PI + theta);
	else
		theta = std::abs(theta);

	// if it hits and not perpendicularly enough, then reflect is needed
	if ( (theta > (PI/2.0)) && (theta < PI-pass_range) ){
		v2f = -un*(a->v * un) + ut*(a->v * ut);
		return true;
	}else
		return false;
}

#ifndef NDEBUG
# define ABSTRACT_BODY { std::cerr << "ERROR: abstract function at " << __FILE__ << ":" << __LINE__ << " has been called!" << std::endl; }
#else
# define ABSTRACT_BODY
#endif

// bouncing function
void au::Automaton::bounce( const util::Vector& newv ){
	ABSTRACT_BODY;
}
// each collision type for active colliders
void au::Automaton::recieve_collision( int& res, Bullet *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Debris *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, ExaminisApicula *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Lagvard *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Laser *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, LightningSpark *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Proton *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, RedHole *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Scared *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Shielded *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Spawned *p ){
	ABSTRACT_BODY;
}
void au::Automaton::recieve_collision( int& res, Spawnship *p ){
	ABSTRACT_BODY;
}
// drawing function
void au::Automaton::specific_draw(){
	ABSTRACT_BODY;
}
// where is the nearest player. itis is an unit vector or zero.
// distance is the distance or REAL_MAX
void au::Automaton::specific_there( const util::Vector& itis, real distance ){
	ABSTRACT_BODY;
}
// process function
void au::Automaton::specific_process( int milliseconds ){
	ABSTRACT_BODY;
}
// dodger parameters
void au::Automaton::specific_dodger_parameters( real& fr, real& fi ) const{
	ABSTRACT_BODY;
}
// dodger function
void au::Automaton::specific_dodger( const util::Vector& f ){
	ABSTRACT_BODY;
}
// collision functions
void au::Automaton::send_collision( int& res, Automaton *a ){
	ABSTRACT_BODY;
}
// scoring points
int au::Automaton::get_killscore() const{
	ABSTRACT_BODY;
	return 0;
}
int au::Automaton::get_hitscore() const{
	ABSTRACT_BODY;
	return 0;
}
// serializes the automaton
void au::Automaton::serialize( std::ostream& out ){
	ABSTRACT_BODY;
}
