// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include "../rc/Constants.hpp"
#include "../util/Random.hpp"
#include "../Main.hpp"
#include "Automaton.hpp"
#include "EnemyScheduler.hpp"

#define MAX_ATTEMPTS 5

// spawn points names
enum spawnpoints {
	NORTH = 0,
	EAST,
	SOUTH,
	WEST,
	NORTHEAST,
	SOUTHEAST,
	SOUTHWEST,
	NORTHWEST,

	DIRECTIONS
};
// fixed spawning points
util::Vector au::EnemyScheduler::points[DIRECTIONS];

// time till next spawn
int au::EnemyScheduler::spawntime;

// is there no spawning process active?
bool au::EnemyScheduler::nospawning;

// which enemies are spawning
bool au::EnemyScheduler::states[Automaton::SPAWNABLE_AMOUNT];
// time left for spawning for each enemy type
int au::EnemyScheduler::spawningtimes[Automaton::SPAWNABLE_AMOUNT];
// spawning points for each enemy type
// bits from lowest order on:
//   NORTH, EAST, SOUTH, WEST, NORTHEAST, SOUTHEAST, SOUTHWEST, NORTHWEST
// or values:
//   NO_POINT for not spawning enemy
//   RAND for random point
//   ALL_POINTS is all possible points
#define NO_POINT    0
#define RAND       -1
#define ALL_POINTS  0x0FF
int au::EnemyScheduler::spawningpoints[Automaton::SPAWNABLE_AMOUNT];

// spawning probability of each enemy
real au::EnemyScheduler::spawning_prob[Automaton::SPAWNABLE_AMOUNT + 1];
// spawning probability of each enemy, added and normalized
real au::EnemyScheduler::spawning_prob_added[Automaton::SPAWNABLE_AMOUNT];
// the enemy is spawned just once at a time
bool au::EnemyScheduler::spawning_single[Automaton::SPAWNABLE_AMOUNT];
// time the spawning process lasts
int  au::EnemyScheduler::spawning_time[Automaton::SPAWNABLE_AMOUNT];
// time between consecutive spawns
int  au::EnemyScheduler::spawning_interval[Automaton::SPAWNABLE_AMOUNT];
// the enemy spawns from random locations
bool au::EnemyScheduler::spawning_rand[Automaton::SPAWNABLE_AMOUNT];

// adds and normalizes the probabilities in [0,1], filling spawning_prob_added
void au::EnemyScheduler::normalize(){
	real mag = 0;
	for( int i=0 ; i<=Automaton::SPAWNABLE_AMOUNT ; ++i )
		mag += spawning_prob[i];

	real added = 0;
	for( int i=0 ; i<Automaton::SPAWNABLE_AMOUNT ; ++i ){
		added += spawning_prob[i];
		spawning_prob_added[i] = (mag>0) ? added/mag : 0;
	}
}
// updates the spawning probabilities
void au::EnemyScheduler::update_prob(){
	spawning_prob[Automaton::EXAMINIS_CODE]    *= rc::examinis_prob_variation();
	spawning_prob[Automaton::LAGVARD_CODE]     *= rc::lagvard_prob_variation();
	spawning_prob[Automaton::REDHOLE_CODE]     *= rc::redhole_prob_variation();
	spawning_prob[Automaton::SCARED_CODE]      *= rc::scared_prob_variation();
	spawning_prob[Automaton::SHIELDED_CODE]    *= rc::shielded_prob_variation();
	spawning_prob[Automaton::SPAWNSHIP_CODE]   *= rc::spawnship_prob_variation();
	spawning_prob[Automaton::SPAWNABLE_AMOUNT] *= rc::no_spawning_prob_variation();
	normalize();
}
// initializes the scheduler [depends on: Constants]
void au::EnemyScheduler::init(){
	// static data
	spawning_prob[Automaton::EXAMINIS_CODE]    = rc::examinis_prob();
	spawning_prob[Automaton::LAGVARD_CODE]     = rc::lagvard_prob();
	spawning_prob[Automaton::REDHOLE_CODE]     = rc::redhole_prob();
	spawning_prob[Automaton::SCARED_CODE]      = rc::scared_prob();
	spawning_prob[Automaton::SHIELDED_CODE]    = rc::shielded_prob();
	spawning_prob[Automaton::SPAWNSHIP_CODE]   = rc::spawnship_prob();
	spawning_prob[Automaton::SPAWNABLE_AMOUNT] = rc::no_spawning_prob();
	normalize();

	spawning_single[Automaton::EXAMINIS_CODE]  = rc::examinis_single();
	spawning_single[Automaton::LAGVARD_CODE]   = rc::lagvard_single();
	spawning_single[Automaton::REDHOLE_CODE]   = rc::redhole_single();
	spawning_single[Automaton::SCARED_CODE]    = rc::scared_single();
	spawning_single[Automaton::SHIELDED_CODE]  = rc::shielded_single();
	spawning_single[Automaton::SPAWNSHIP_CODE] = rc::spawnship_single();

	spawning_time[Automaton::EXAMINIS_CODE]  = rc::examinis_time();
	spawning_time[Automaton::LAGVARD_CODE]   = rc::lagvard_time();
	spawning_time[Automaton::REDHOLE_CODE]   = rc::redhole_time();
	spawning_time[Automaton::SCARED_CODE]    = rc::scared_time();
	spawning_time[Automaton::SHIELDED_CODE]  = rc::shielded_time();
	spawning_time[Automaton::SPAWNSHIP_CODE] = rc::spawnship_time();

	spawning_interval[Automaton::EXAMINIS_CODE]  = rc::examinis_int();
	spawning_interval[Automaton::LAGVARD_CODE]   = rc::lagvard_int();
	spawning_interval[Automaton::REDHOLE_CODE]   = rc::redhole_int();
	spawning_interval[Automaton::SCARED_CODE]    = rc::scared_int();
	spawning_interval[Automaton::SHIELDED_CODE]  = rc::shielded_int();
	spawning_interval[Automaton::SPAWNSHIP_CODE] = rc::spawnship_int();

	spawning_rand[Automaton::EXAMINIS_CODE]  = rc::examinis_rand();
	spawning_rand[Automaton::LAGVARD_CODE]   = rc::lagvard_rand();
	spawning_rand[Automaton::REDHOLE_CODE]   = rc::redhole_rand();
	spawning_rand[Automaton::SCARED_CODE]    = rc::scared_rand();
	spawning_rand[Automaton::SHIELDED_CODE]  = rc::shielded_rand();
	spawning_rand[Automaton::SPAWNSHIP_CODE] = rc::spawnship_rand();

	points[NORTH]     = util::Vector( rc::ring_middlex(),     rc::ring_top() );
	points[EAST]      = util::Vector(   rc::ring_right(), rc::ring_middley() );
	points[SOUTH]     = util::Vector( rc::ring_middlex(),  rc::ring_bottom() );
	points[WEST]      = util::Vector(    rc::ring_left(), rc::ring_middley() );
	points[NORTHEAST] = util::Vector(   rc::ring_right(),     rc::ring_top() );
	points[SOUTHEAST] = util::Vector(   rc::ring_right(),  rc::ring_bottom() );
	points[SOUTHWEST] = util::Vector(    rc::ring_left(),  rc::ring_bottom() );
	points[NORTHWEST] = util::Vector(    rc::ring_left(),     rc::ring_top() );

	// setting that nothing is being done
	stop();
}
// stops present spawning
void au::EnemyScheduler::stop(){
	// noone is spawning
	nospawning = true;
	for( int i=0 ; i<Automaton::SPAWNABLE_AMOUNT ; ++i ){
		states[i] = false;
		spawningpoints[i] = NO_POINT;
	}

	// we will spawn after a peace time
	spawntime = rc::peace_time();
}
// updates spawning
void au::EnemyScheduler::update( int milliseconds ){
	// tick time, taking into account the game's present activity
	spawntime -= static_cast<int>(milliseconds * Main::activity());

	// if the time for spawn has arrived, or not enough enemies alive
	if ( (spawntime<0) || (nospawning && (Automaton::enemies_amount()<rc::enemy_threshold())) ){
		spawntime = rc::peace_time();

		// spawning a random enemy
		real ran = util::Random::uget();

		for( int i=0 ; i<Automaton::SPAWNABLE_AMOUNT ; ++i ){
			if ( ran < spawning_prob_added[i] ){
				// we are now spawning something
				nospawning = false;
				// setting spawn information
				states[i] = true;
				spawningtimes[i] = spawning_time[i];
				if ( spawning_rand[i] )
					spawningpoints[i] = RAND;
				else{
					// computing the 4 (at most) spawning points for the future enemies
					//  <0: NORTH, EAST, SOUTH, WEST
					//  >0: NORTHEAST, SOUTHEAST, SOUTHWEST, NORTHWEST
					int p = (util::Random::get()<0.0) ? 0 : 4;
					do{
						if ( Automaton::player_free_area( points[p], rc::big_spawning_area_r() ) )
							spawningpoints[i] |= (0x01<<p);
						++p;
					}while( p & 0x03 );
				}
				break;
			}
		}
	}

	// spawning what is spawning
	if ( ! nospawning ){
		for( int i=0 ; i<Automaton::SPAWNABLE_AMOUNT ; ++i ){
			// if i is spawning
			if ( states[i] ){
				// checking if a spawn is needed now
				bool dospawn;
				if ( spawning_single[i] )
					dospawn = true;
				else{
					// taking into account the intervals
					int interval = spawning_interval[i];
					int _time = spawningtimes[i];

					spawningtimes[i] -= milliseconds;

					dospawn = (_time/interval) != (spawningtimes[i]/interval);
				}
				// doing a spawn
				if ( dospawn ){
					if ( spawningpoints[i] == RAND ){
						// doing a random point spawn
						int j = MAX_ATTEMPTS;
						bool fr;
						util::Vector xy;
						do{
							xy = util::Vector( util::Random::get( rc::ring_left(), rc::ring_right() ),
							                   util::Random::get( rc::ring_bottom(), rc::ring_top() ) );
							--j;
							fr = Automaton::player_free_area( xy, rc::big_spawning_area_r() );
						}while( j && !fr );

						if ( fr )
							Automaton::create_enemy( i, xy );
					}else{
						// doing a fixed point spawn
						for( int j=0 ; j<DIRECTIONS ; ++j ){
							if ( (spawningpoints[i] & (0x01<<j)) && Automaton::player_free_area( points[j], rc::small_spawning_area_r() ) )
								Automaton::create_enemy( i, points[j] );
						}
					}
				}
				// cleaning state if spawning is done
				if ( spawning_single[i] || (spawningtimes[i]<0) ){
					states[i] = false;
					spawningpoints[i] = NO_POINT;

					nospawning = true;
					for( int j=0 ; nospawning && (j<Automaton::SPAWNABLE_AMOUNT) ; ++j )
						nospawning = ! states[j];
				}
			}
		}
	}
}
// serializes the enemy scheduler
void au::EnemyScheduler::serialize( std::ostream& out ){
	// saving time
	out << spawntime << std::endl;
	// saving state
	out << nospawning;
	if ( ! nospawning ){
		for( int i=0 ; i<Automaton::SPAWNABLE_AMOUNT ; ++i )
			out << " " << states[i] << " " << spawningtimes[i] << " " << spawningpoints[i];
	}
	out << std::endl;
	// saving probabilities
	for( int i=0 ; i<=Automaton::SPAWNABLE_AMOUNT ; ++i ){
		if ( i )
			out << " ";
		out << spawning_prob[i];
	}
	out << std::endl;
}
// unserializes the enemy scheduler
void au::EnemyScheduler::unserialize( std::istream& in ){
	// loading time
	in >> spawntime;
	if ( spawntime < 0 )
		throw Main::UNSERIAL_SPAWNTIME;
	// loading state
	in >> nospawning;
	if ( ! nospawning ){
		for( int i=0 ; i<Automaton::SPAWNABLE_AMOUNT ; ++i ){
			in >> states[i] >> spawningtimes[i] >> spawningpoints[i];
			if ( states[i] && (spawningtimes[i] < 0) )
				throw Main::UNSERIAL_SPAWNINGTIMES;
			if ( ((spawningpoints[i] != RAND) && (spawningpoints[i] & ~ALL_POINTS))
			     || (!states[i] && (spawningpoints[i] != NO_POINT)) )
				throw Main::UNSERIAL_SPAWNINGPOINTS;
		}
	}else{
		for( int i=0 ; i<Automaton::SPAWNABLE_AMOUNT ; ++i ){
			states[i] = false;
			spawningpoints[i] = NO_POINT;
		}
	}
	// loading probabilities
	bool found = false;
	for( int i=0 ; i<=Automaton::SPAWNABLE_AMOUNT ; ++i ){
		in >> spawning_prob[i];
		if ( spawning_prob[i] < 0 )
			throw Main::UNSERIAL_SPAWNING_PROB;
		if ( spawning_prob[i] > 0 )
			found = true;
	}
	if ( ! found )
		throw Main::UNSERIAL_SPAWNING_PROB;
	normalize();
}
