// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include "../rc/Constants.hpp"
#include "../vis/Draw.hpp"
#include "../Config.hpp"
#include "../Main.hpp"
#include "Automaton.hpp"
#include "Grid.hpp"

// the grid
util::Matrix<au::Spring*> au::Grid::grid;
int au::Grid::rows;
int au::Grid::cols;

// initializes the grid [depends on: Config]
void au::Grid::init(){
	rows = 0;
	cols = 0;

	reinit();
}
// reinitializes the grid, call it when config changes
void au::Grid::reinit(){
	if ( Config::get_grid() ){
		// if there must be a grid
		if ( ! grid.empty() ){
			// if we already had one
			if ( (rows == Config::get_gridrows()) && (cols == Config::get_gridcols()) )
				return; // if it was ok, do nothing
			deletegrid();
		}

		rows = Config::get_gridrows();
		cols = Config::get_gridcols();

		// creating grid
		grid = util::Matrix<Spring*>( rows-1, cols-1 );

		for( int i=0 ; i<grid.get_rows() ; ++i ){
			for( int j=0 ; j<grid.get_cols() ; ++j ){
				grid[i][j] = Automaton::create_spring(
					(j+1)*rc::width()/cols + rc::ring_x0(),
					(i+1)*rc::height()/rows + rc::ring_y0() );
			}
		}
	}else{
		// if there must be no grid
		if ( ! grid.empty() )
			deletegrid(); // erase it if we had one
		rows = 0;
		cols = 0;
	}
}
// ends the grid
void au::Grid::finalize(){
	if ( ! grid.empty() )
		deletegrid();
}
// deletes the grid
void au::Grid::deletegrid(){
	for( int i=0 ; i<grid.get_rows() ; ++i ){
		for( int j=0 ; j<grid.get_cols() ; ++j ){
			delete grid[i][j];
		}
	}
	grid.clear();
}

// serializes the grid
void au::Grid::serialize( std::ostream& out ){
	out << rows << " " << cols << std::endl;
	if ( ! grid.empty() ){
		for( int i=0 ; i<grid.get_rows() ; ++i ){
			for( int j=0 ; j<grid.get_cols() ; ++j ){
				grid[i][j]->serialize( out );
			}
		}
	}
}
// postoperations when the serialization has finished
void au::Grid::end_serialize(){
	if ( ! grid.empty() ){
		for( int i=0 ; i<grid.get_rows() ; ++i ){
			for( int j=0 ; j<grid.get_cols() ; ++j ){
				grid[i][j]->end_serialize();
			}
		}
	}
}
// unserializes the grid
void au::Grid::unserialize( std::istream& in ){
	in >> rows >> cols;
	if ( (rows || cols) && ((rows<ROWS_MIN) || (cols<COLS_MIN)) )
		throw Main::UNSERIAL_GRID_DIM;
	if ( rows ){
		grid = util::Matrix<Spring*>( rows-1, cols-1 );
		for( int i=0 ; i<grid.get_rows() ; ++i ){
			for( int j=0 ; j<grid.get_cols() ; ++j ){
				grid[i][j] = Spring::unserialize( in );
			}
		}
	}
	// syncing present state with its pertinent config
	Config::set_grid( rows );
	Config::set_gridrows( rows );
	Config::set_gridcols( cols );
}

// drawing function
void au::Grid::draw(){
	vis::Draw::reset();

	if ( ! grid.empty() ){
		vis::Draw::set_thin_lines();
		if ( Config::get_gridbezier() )
			drawbezier();
		else
			drawlines();
	}
}
// draw grid using lines
void au::Grid::drawlines(){
	// drawing horizontal lines
	for( int i=0 ; i<grid.get_rows() ; ++i ){
		vis::Draw::start_line_strip();
			real auxY = (i+1)*rc::height()/rows + rc::ring_y0(); // current line's Y position

			// fixed fake starting point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][0]->get_color_alpha() );
			vis::Draw::add_point( rc::ring_x0(), auxY );

			// drawing all real points in the line
			for( int j=0 ; j<grid.get_cols() ; ++j ){
				vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
				vis::Draw::add_point( grid[i][j]->get_xy() );
			}

			// fixed fake ending point
			vis::Draw::add_point( rc::ring_x1(), auxY );
		vis::Draw::end_line_strip();
	}
	// drawing vertical lines
	for( int j=0 ; j<grid.get_cols() ; ++j ){
		vis::Draw::start_line_strip();
			real auxX = (j+1)*rc::width()/cols + rc::ring_x0(); // current line's X position

			// fixed fake starting point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[0][j]->get_color_alpha() );
			vis::Draw::add_point( auxX, rc::ring_y0() );

			// drawing all real points in the line
			for( int i=0 ; i<grid.get_rows() ; ++i ){
				vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
				vis::Draw::add_point( grid[i][j]->get_xy() );
			}

			// fixed fake ending point
			vis::Draw::add_point( auxX, rc::ring_y1() );
		vis::Draw::end_line_strip();
	}
}
void au::Grid::drawbezier(){
	// drawing horizontal lines
	real radX = rc::grid_control_point_r() * rc::width()/cols; // radius of the control points
	for( int i=0 ; i<grid.get_rows() ; ++i ){
		vis::Draw::start_line_strip();
			real auxY = (i+1)*rc::height()/rows + rc::ring_y0(); // current line's Y position
			util::Vector pA;
			util::Vector pB;
			util::Vector paux;

			// fixed fake starting point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][0]->get_color_alpha() );
			vis::Draw::add_point( rc::ring_x0(), auxY );

			calcPoints( util::Vector(rc::ring_x0(), auxY), grid[i][0]->get_xy(), grid[i][1]->get_xy(), radX, pA, pB );

			// drawing first real point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][0]->get_color_alpha() );
			vis::Draw::add_curve( rc::ring_x0()+radX, auxY, pA, grid[i][0]->get_xy() );

			// drawing all real points in the line, except first and last
			int j;
			for( j=1 ; j<grid.get_cols()-1 ; ++j ){
				paux = pB;
				calcPoints( grid[i][j-1]->get_xy(), grid[i][j]->get_xy(), grid[i][j+1]->get_xy(), radX, pA, pB );

				vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
				vis::Draw::add_curve( paux, pA, grid[i][j]->get_xy() );
			}

			paux = pB;
			calcPoints( grid[i][j-1]->get_xy(), grid[i][j]->get_xy(), util::Vector(rc::ring_x1(), auxY), radX, pA, pB );

			// drawing last real point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
			vis::Draw::add_curve( paux, pA, grid[i][j]->get_xy() );

			// fixed fake ending point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
			vis::Draw::add_curve( pB, rc::ring_x1()-radX, auxY, rc::ring_x1(), auxY );
		vis::Draw::end_line_strip();
	}
	// drawing vertical lines
	real radY = rc::grid_control_point_r() * rc::height()/rows; // radius of the control points
	for( int j=0 ; j<grid.get_cols() ; ++j ){
		vis::Draw::start_line_strip();
			real auxX = (j+1)*rc::width()/cols + rc::ring_x0(); // current line's X position
			util::Vector pA;
			util::Vector pB;
			util::Vector paux;

			// fixed fake starting point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[0][j]->get_color_alpha() );
			vis::Draw::add_point( auxX, rc::ring_y0() );

			calcPoints( util::Vector(auxX, rc::ring_y0()), grid[0][j]->get_xy(), grid[1][j]->get_xy(), radY, pA, pB );

			// drawing first real point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[0][j]->get_color_alpha() );
			vis::Draw::add_curve( auxX, rc::ring_y0()+radY, pA, grid[0][j]->get_xy() );

			// drawing all real points in the line, except first and last
			int i;
			for( i=1 ; i<grid.get_rows()-1 ; ++i ){
				paux = pB;
				calcPoints( grid[i-1][j]->get_xy(), grid[i][j]->get_xy(), grid[i+1][j]->get_xy(), radY, pA, pB );

				vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
				vis::Draw::add_curve( paux, pA, grid[i][j]->get_xy() );
			}

			paux = pB;
			calcPoints( grid[i-1][j]->get_xy(), grid[i][j]->get_xy(), util::Vector(auxX, rc::ring_y1()), radY, pA, pB );

			// drawing last real point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
			vis::Draw::add_curve( paux, pA, grid[i][j]->get_xy() );

			// fixed fake ending point
			vis::Draw::set_color( rc::spring_c_r(), rc::spring_c_g(), rc::spring_c_b(), grid[i][j]->get_color_alpha() );
			vis::Draw::add_curve( pB, auxX, rc::ring_y1()-radY, auxX, rc::ring_y1() );
		vis::Draw::end_line_strip();
	}
}
// compute bezier control points for act
void au::Grid::calcPoints( const util::Vector& prev,
                           const util::Vector& act,
                           const util::Vector& next,
                           const real rad,
                           util::Vector& actA,
                           util::Vector& actB ){
	util::Vector vec = (next - prev).magnitude(rad);
	actA = act - vec;
	actB = act + vec;
}
