// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <algorithm>
#include "../vis/Draw.hpp"
#include "../Audio.hpp"
#include "../Main.hpp"
#include "Player.hpp"

// ctor
au::Player::Player( int id, const util::Vector& xy, int t, int ltime, int ftime, real angle, bool imm ) :
 Automaton( xy, util::Vector(0.0,0.0), rc::player_collision_r(), t ),
 id( id ),
 input( Input::get_input(id) ),
 ltime( ltime ),
 ftime( ftime ),
 angle( angle ),
 imm( imm )
{
	start_activecollider( PLAYER_COL );
	start_gravitylistener();
	start_processable();
	start_visible( PLAYER_PRIO );
	start_player();

	Main::add_player( id );
}
// dtor
au::Player::~Player(){
	Main::remove_player( id );
}
// collision checking
int au::Player::generic_collision(){
	if ( imm || (t < rc::player_imm_t()) )
		return PASSIVE_DIES;
	else{
		Audio::playSound( Audio::SOUND_PLAYER_DEATH );
		Automaton::create_blast( xy );
		return ACTIVE_DIES;
	}
}
void au::Player::recieve_collision( int& res, ExaminisApicula *p ){
	res = generic_collision();
}
void au::Player::recieve_collision( int& res, Lagvard *p ){
	res = generic_collision();
}
void au::Player::recieve_collision( int& res, Proton *p ){
	res = generic_collision();
}
void au::Player::recieve_collision( int& res, RedHole *p ){
	res = generic_collision();
}
void au::Player::recieve_collision( int& res, Scared *p ){
	res = generic_collision();
}
void au::Player::recieve_collision( int& res, Shielded *p ){
	res = generic_collision();
}
void au::Player::recieve_collision( int& res, Spawned *p ){
	res = generic_collision();
}
void au::Player::recieve_collision( int& res, Spawnship *p ){
	res = generic_collision();
}
// serializes the automaton
void au::Player::serialize( std::ostream& out ){
	out << Automaton::PLAYER_SERIAL
	    << " " << xy << " " << t
	    << " " << id << " " << ltime << " " << ftime << " " << angle << " " << imm << std::endl;
}
// unserializes the automaton
void au::Player::unserialize( std::istream& in ){
	util::Vector xy;
	int t;
	int id;
	int ltime;
	int ftime;
	real angle;
	bool imm;

	in >> xy >> t
	   >> id >> ltime >> ftime >> angle >> imm;
	if ( t < 0 )
		throw Main::UNSERIAL_PLAYER_T;
	if ( (id < 0) || (id >= static_cast<int>(Input::get_inputs().size())) )
		throw Main::UNSERIAL_PLAYER_ID;
	if ( ltime < 0 )
		throw Main::UNSERIAL_PLAYER_LTIME;
	if ( ftime < 0 )
		throw Main::UNSERIAL_PLAYER_FTIME;

	new Player( id, xy, t, ltime, ftime, angle, imm );
}
// drawing function
void au::Player::specific_draw(){
	vis::Draw::translate( xy );
	// draw player's immunity aura
	if ( imm )
		vis::Draw::play_record( vis::PLAYER_ID0 );
	else if ( t < rc::player_imm_t() ){
		int aux = t - (rc::player_imm_t() - rc::player_defense_blinking_time());
		if ( (aux < 0) || ((aux/rc::player_defense_blink_time()) & 1) )
			vis::Draw::play_record( vis::PLAYER_ID0 );
	}
	// draw player
	vis::Draw::rotate( angle ); // look to where user is/was heading
	vis::Draw::play_record( vis::PLAYER_ID );
}

// processing function
void au::Player::specific_process( int milliseconds ){
	v = input->get_move() * rc::player_s();
	xy += v*milliseconds + g*(milliseconds*rc::non_inertial_time_multiplier());

	// checking ring with a fake radius
	real _r = r;
	r = rc::player_ring_r();
	checkRing_stop();
	r = _r;

	// littering
	if ( input->get_move() == util::Vector(0.0, 0.0) )
		ltime = rc::player_litter_int();
	else{
		// computing what direction is the ship looking at
		angle = input->get_move().angle();

		ltime += milliseconds;
		if ( ltime >= rc::player_litter_int() ){
			ltime = 0;
			Automaton::create_debris( xy, rc::player_litter_r(), angle-PI, rc::player_litter_phi() );
		}
	}

	// firing
	if ( input->get_fire() == util::Vector(0.0, 0.0) )
		ftime = Automaton::shot_rate();
	else{
		ftime += milliseconds;
		if ( ftime >= Automaton::shot_rate() ){
			ftime = 0;
			Automaton::create_shot( xy, input->get_fire(), rc::player_shoot_r() );
		}
	}

	// bombing
	if ( input->get_bomb() )
		Main::bombing( xy );

	// checking immortality
	if ( input->get_imm() )
		imm = ! imm;
}
