// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <algorithm>
#include "../vis/Draw.hpp"
#include "../Audio.hpp"
#include "../Main.hpp"
#include "RedHole.hpp"

// ctor
au::RedHole::RedHole( const util::Vector& xy, const util::Vector& v, real m, real criticalmass, bool sleep, int feed, real damage, real massincr, bool play ) :
 Automaton( xy, v, rc::redhole_r()*rc::redhole_birth_scale() ),
 criticalmass( criticalmass ),
 sleep( sleep ),
 feed( feed ),
 scale( rc::redhole_birth_scale() ),
 damage( damage ),
 massincr( massincr ),
 dospawn( false )
{
	this->m = m;
	rescale();

	start_activecollider( REDHOLE_COL );
	start_passivecollider( PLAYER_COL | REDHOLE_COL );
	start_enemy( play );
	start_gravitylistener();
	start_processable();
	start_visible( REDHOLE_PRIO );

	if ( ! sleep ){
		Audio::playSound( Audio::SOUND_REDHOLE_DRAG, true );
		start_gravityprovider();
	}
}
// dtor
au::RedHole::~RedHole(){
	if ( ! sleep )
		Audio::stopSound( Audio::SOUND_REDHOLE_DRAG );

	if ( dolitter )
		Automaton::create_debris( xy, r, rc::redhole_litter_c_r(), rc::redhole_litter_c_g(), rc::redhole_litter_c_b() );

	// spawning protons
	if ( dospawn )
		Automaton::create_protons( xy, r, m );
}
// updates the scale
void au::RedHole::rescale(){
	if ( ! sleep ){
		scale = std::max( 1.0, (m-damage)/(2.0*rc::redhole_mass()) );
		r = rc::redhole_r() * scale;
	}
}

// collision checking
int au::RedHole::shot_collision(){
	feed = 0;

	if ( sleep ){
		Audio::playSound( Audio::SOUND_REDHOLE_DRAG );
		sleep = false;
		start_gravityprovider();
	}

	damage += rc::redhole_mass() / rc::redhole_lifes();
	rescale();

	Main::add_score( get_hitscore() );

	// check if it has killed us
	if ( damage > m ){
		dospawn = true;
		Main::add_score( get_killscore() );
		return BOTH_DIE;
	}else
		return PASSIVE_DIES;
}
int au::RedHole::enemy_collision( Automaton *p, bool debris ){
	if ( ! debris )
		feed = 0;

	if ( sleep )
		return NO_DEATHS;
	else{
		m += p->get_mass();
		massincr += p->get_mass();
		rescale();

		p->no_litter();

		return PASSIVE_DIES;
	}
}
// each collision type for active colliders
void au::RedHole::recieve_collision( int& res, Bullet *p ){
	res = shot_collision();
}
void au::RedHole::recieve_collision( int& res, Debris *p ){
	res = enemy_collision( p, true );
}
void au::RedHole::recieve_collision( int& res, Lagvard *p ){
	res = enemy_collision( p );
}
void au::RedHole::recieve_collision( int& res, Laser *p ){
	res = shot_collision();
}
void au::RedHole::recieve_collision( int& res, LightningSpark *p ){
	res = shot_collision();
}
void au::RedHole::recieve_collision( int& res, RedHole *p ){
	feed = 0;
	p->feed = 0;

	ellastic_collision( p, m, p->m, rc::redhole_s_lose_c() );
	avoid_collision( p );

	res = NO_DEATHS;
}
void au::RedHole::recieve_collision( int& res, Scared *p ){
	res = enemy_collision( p );
}
void au::RedHole::recieve_collision( int& res, Spawned *p ){
	res = enemy_collision( p );
}
void au::RedHole::recieve_collision( int& res, Spawnship *p ){
	res = enemy_collision( p );
}
// serializes the automaton
void au::RedHole::serialize( std::ostream& out ){
	out << Automaton::REDHOLE_SERIAL
	    << " " << xy << " " << v << " " << m
	    << " " << criticalmass << " " << sleep << " " << feed << " " << damage << " " << massincr << std::endl;
}
// unserializes the automaton
void au::RedHole::unserialize( std::istream& in ){
	util::Vector xy;
	util::Vector v;
	real m;
	real criticalmass;
	bool sleep;
	int feed;
	real damage;
	real massincr;

	in >> xy >> v >> m
	   >> criticalmass >> sleep >> feed >> damage >> massincr;
	if ( (m < 0) || (m >= criticalmass) )
		throw Main::UNSERIAL_REDHOLE_M;
	if ( feed < 0 )
		throw Main::UNSERIAL_REDHOLE_FEED;
	if ( (damage < 0) || (damage > m) )
		throw Main::UNSERIAL_REDHOLE_DAMAGE;
	if ( (massincr < 0) || (massincr >= rc::redhole_mass_i()) )
		throw Main::UNSERIAL_REDHOLE_MASSINCR;

	new RedHole( xy, v, m, criticalmass, sleep, feed, damage, massincr, false );
}
// drawing function
void au::RedHole::specific_draw(){
	// draw red hole
	vis::Draw::set_color( rc::redhole_c_r(), rc::redhole_c_g(), rc::redhole_c_b() );
	vis::Draw::translate( xy );
	vis::Draw::scale( scale );
	vis::Draw::play_record( vis::REDHOLE_ID );

	if ( feed < rc::redhole_t() ){
		// draw inner red hole
		vis::Draw::set_color( rc::redhole_inner_c_r(), rc::redhole_inner_c_g(), rc::redhole_inner_c_b() );
		vis::Draw::scale( rc::redhole_inner_scale() );
		vis::Draw::play_record( vis::REDHOLE_ID );
	}
}
// process function
void au::RedHole::specific_process( int milliseconds ){
	feed += milliseconds;

	xy += v*milliseconds + g*(milliseconds*milliseconds/2.0);

	v += g * milliseconds;

	checkRing_bounce( rc::redhole_s_lose_c() );

	if ( m >= criticalmass ){
		// if critical mass reached, die
		dospawn = true;
		add_zombie();
		Automaton::create_blast( xy );
	}else if ( massincr >= rc::redhole_mass_i() ){
		// if another mass increment, aproaching critical mass so warn user
		massincr = 0.0;
		Automaton::create_blast( xy );
	}
}
// collision functions
void au::RedHole::send_collision( int& res, Automaton *a ){
	a->recieve_collision( res, this );
}
// scoring points
int au::RedHole::get_killscore() const{
	return rc::redhole_kill_score();
}
int au::RedHole::get_hitscore() const{
	return rc::redhole_hit_score();
}
