// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cmath>
#include "../vis/Draw.hpp"
#include "../Main.hpp"
#include "Shielded.hpp"

// ctor
au::Shielded::Shielded( const util::Vector& xy, const util::Vector& v, int lifes, real phase, bool play ) :
 Automaton( xy, v, rc::shielded_r() ),
 lifes( lifes ),
 phase( phase )
{
	start_activecollider( SHIELDED_COL );
	start_passivecollider( PLAYER_COL | SHIELDED_COL );
	start_enemy( play );
	start_processable();
	start_visible( SHIELDED_PRIO );
}
// dtor
au::Shielded::~Shielded(){
	if ( dolitter ){
		Automaton::create_debris( xy, rc::shielded_litter_r(),
		                          rc::shielded_litter_c_r(), rc::shielded_litter_c_g(), rc::shielded_litter_c_b() );
		Automaton::create_debris( xy, rc::shielded_litter_inner_r(),
		                          rc::shielded_litter_c_r(), rc::shielded_litter_c_g(), rc::shielded_litter_c_b() );
	}
}
// collision checking
int au::Shielded::shot_collision( Automaton *p ){
	if ( intersection_mag(p) > rc::shielded_w() ){
		// a shot is inside the shield
		if ( p->intersects_circle( xy, rc::shielded_inner_r() ) ){
			// it is even inside the shield's heart
			Main::add_score( get_hitscore() );
			if ( --lifes )
				return PASSIVE_DIES;
			else{
				Main::add_score( get_killscore() );
				return BOTH_DIE;
			}
		}else
			return NO_DEATHS;
	}else{
		// reflecting shots
		util::Vector v2;
		if ( reflect_collision( p, rc::shielded_pass_range(), v2 ) )
			p->bounce( v2 );

		return NO_DEATHS;
	}
}
// each collision type for active colliders
void au::Shielded::recieve_collision( int& res, Bullet *p ){
	res = shot_collision( p );
}
void au::Shielded::recieve_collision( int& res, Laser *p ){
	res = shot_collision( p );
}
void au::Shielded::recieve_collision( int& res, LightningSpark *p ){
	res = shot_collision( p );
}
void au::Shielded::recieve_collision( int& res, Shielded *p ){
	ellastic_collision( p, 1.0, 1.0, rc::shielded_s_lose_c() );
	avoid_collision( p );
	res = NO_DEATHS;
}
// serializes the automaton
void au::Shielded::serialize( std::ostream& out ){
	out << Automaton::SHIELDED_SERIAL
	    << " " << xy << " " << v
	    << " " << lifes << " " << phase << std::endl;
}
// unserializes the automaton
void au::Shielded::unserialize( std::istream& in ){
	util::Vector xy;
	util::Vector v;
	int lifes;
	real phase;

	in >> xy >> v
	   >> lifes >> phase;
	if ( lifes <= 0 )
		throw Main::UNSERIAL_SHIELDED;

	new Shielded( xy, v, lifes, phase, false );
}
// drawing function
void au::Shielded::specific_draw(){
	// inner saw
	vis::Draw::set_color( rc::shielded_inner_c_r(), rc::shielded_inner_c_g(), rc::shielded_inner_c_b(),
	                      static_cast<real>(lifes)/rc::shielded_lifes() );
	vis::Draw::translate( xy );
	vis::Draw::rotate( phase );
	vis::Draw::play_record( vis::SHIELDED_ID );

	vis::Draw::reset();

	// outter saw
	vis::Draw::set_color( rc::shielded_c_r(), rc::shielded_c_g(), rc::shielded_c_b() );
	vis::Draw::translate( xy );
	vis::Draw::rotate( -phase );
	vis::Draw::scale( rc::shielded_r()/rc::shielded_inner_r(), -rc::shielded_r()/rc::shielded_inner_r() );
	vis::Draw::play_record( vis::SHIELDED_ID );
}
// process function
void au::Shielded::specific_process( int milliseconds ){
	phase = std::fmod( phase+(rc::shielded_as()*milliseconds), static_cast<real>(2.0*PI) );

	xy += v*milliseconds;

	checkRing_bounce( rc::shielded_s_lose_c() );
}
// collision functions
void au::Shielded::send_collision( int& res, Automaton *a ){
	a->recieve_collision( res, this );
}
// scoring points
int au::Shielded::get_killscore() const{
	return rc::shielded_kill_score();
}
int au::Shielded::get_hitscore() const{
	return rc::shielded_hit_score();
}
