// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cmath>
#include <cstdlib>
#include "../rc/Constants.hpp"
#include "../vis/Draw.hpp"
#include "../Input.hpp"
#include "../Audio.hpp"
#include "Menu_Root.hpp"
#include "Menu_Config.hpp"
#include "Menu_Controls.hpp"
#include "Menu.hpp"

#define CAPTURING_JOY_THRESH 10922

// menu has been activated
bool gui::Menu::on;
// time since last player interaction
int gui::Menu::time;

// menu drawing has to be updated
bool gui::Menu::updated;

// active pane
int gui::Menu::pane;
// available panes
gui::Menu_Component *gui::Menu::panes[] = { NULL, NULL, NULL };
// pane's renders
vis::Font *gui::Menu::renders[] = { NULL, NULL, NULL };

// capturing?
bool gui::Menu::capturing;
// present capturing process type
int  gui::Menu::capture_type;
// pane to notify
int  gui::Menu::capture_notifywho;
// active joystick axis capture data
int  gui::Menu::capture_jwhich;
int  gui::Menu::capture_jaxis;
int  gui::Menu::capture_jvalue;
bool gui::Menu::capture_jcount;
int  gui::Menu::capture_jtime;

// initializes the menu [depends on: -]
void gui::Menu::init(){
	on = false;
	panes[ROOT_PANE] = new Menu_Root();
	panes[CONFIG_PANE] = new Menu_Config();
	panes[CONTROLS_PANE] = new Menu_Controls();

	updated = false;

	capturing = false;
}
// ends the menu
void gui::Menu::finalize(){
	// deleting panes
	if ( panes[ROOT_PANE] != NULL )
		delete panes[ROOT_PANE];
	if ( panes[CONFIG_PANE] != NULL )
		delete panes[CONFIG_PANE];
	if ( panes[CONTROLS_PANE] != NULL )
		delete panes[CONTROLS_PANE];

	// deleting panes' renders
	if ( renders[ROOT_PANE] != NULL )
		delete renders[ROOT_PANE];
	if ( renders[CONFIG_PANE] != NULL )
		delete renders[CONFIG_PANE];
	if ( renders[CONTROLS_PANE] != NULL )
		delete renders[CONTROLS_PANE];
}
// initialize drawing, deletes and NULLifies the renders
void gui::Menu::initdrawing(){
	if ( renders[ROOT_PANE] != NULL ){
		delete renders[ROOT_PANE];
		renders[ROOT_PANE] = NULL;
	}
	if ( renders[CONFIG_PANE] != NULL ){
		delete renders[CONFIG_PANE];
		renders[CONFIG_PANE] = NULL;
	}
	if ( renders[CONTROLS_PANE] != NULL ){
		delete renders[CONTROLS_PANE];
		renders[CONTROLS_PANE] = NULL;
	}

	updated = true;
}
// menu has been activated?
bool gui::Menu::is_on(){
	return on;
}
// turn on the menu
void gui::Menu::turnon(){
	on = true;
	time = rc::menu_response_time();

	pane = ROOT_PANE;
	panes[pane]->start( NO_PANE );

	updated = true;

	Audio::stop();
}
// turn off the menu
void gui::Menu::turnoff(){
	on = false;

	Audio::reinit();
}
// updates the menu
void gui::Menu::update(){
	updated = true;
}
// starts the who pane_id pane or turns off the menu if who is NO_PANE
void gui::Menu::start( int who ){
	if ( who == NO_PANE )
		turnoff();
	else{
		int from = pane;

		pane = who;
		panes[pane]->start( from );

		updated = true;
	}
}
// starts a capture_t type capture and will notify pane notifywho
void gui::Menu::capture( int type, int notifywho ){
	capturing = true;
	capture_type = type;
	capture_notifywho = notifywho;

	// reseting joystick axis information
	capture_jvalue = 0;
	capture_jcount = false;
	// let the joystick return to a resting position before capture starts
	capture_jtime  = -rc::joystick_capturing_time();
}
// key event capturing function
void gui::Menu::captureKey( SDLKey key ){
	if ( ! capturing )
		return;

	// if we are capturing the joystick
	if ( (capture_type==CAPTURE_JOYSTICK_BUTTON) || (capture_type==CAPTURE_JOYSTICK_AXIS) ){
		// and someone presses escape, cancel capture
		if ( Input::get_agg_escape() ){
			panes[capture_notifywho]->notify( false, SDLK_UNKNOWN, -1, -1, false );
			capture_type = NO_CAPTURE;
		}
	}

	// if we are capturing the keyboard, we are done
	if ( capture_type == CAPTURE_KEY ){
		panes[capture_notifywho]->notify( true, key, -1, -1, false );
		capture_type = NO_CAPTURE;
	}
}
// button event capturing function
void gui::Menu::captureButton( int which, int button ){
	if ( ! capturing )
		return;

	// if we are capturing the joystick buttons, we are done
	if ( capture_type == CAPTURE_JOYSTICK_BUTTON ){
		panes[capture_notifywho]->notify( true, SDLK_UNKNOWN, which, button, false );
		capture_type = NO_CAPTURE;
	}
}
// axis event capturing function
void gui::Menu::captureAxis( int which, int axis, int value ){
	if ( ! capturing )
		return;

	// if we are capturing the joystick axis, we are not done yet
	if ( capture_type == CAPTURE_JOYSTICK_AXIS ){
		// if axis movement is greater than the treshold
		if ( std::abs(value) > CAPTURING_JOY_THRESH ){
			// and greater than previous recorded value
			if ( std::abs(value) >= std::abs(capture_jvalue) ){
				// we have a candidate
				capture_jwhich = which;
				capture_jaxis  = axis;
				capture_jvalue = value;
			}
			capture_jcount = true;
		}
	}
}
// color callback
void gui::Menu::color( int part ){
	if ( part == panes[pane]->get_line() )
		vis::Draw::set_color( rc::menu_active_c_r(), rc::menu_active_c_g(), rc::menu_active_c_b() );
	else
		vis::Draw::set_color( rc::menu_normal_c_r(), rc::menu_normal_c_g(), rc::menu_normal_c_b() );
}
// drawing function
void gui::Menu::draw(){
	vis::Draw::reset();

	if ( on ){
		if ( updated ){
			updated = false;

			// updating the pane
			if ( renders[pane] != NULL )
				delete renders[pane];
			renders[pane] = new vis::Font( vis::Font::MENU_TEXT_FRAMED, panes[pane]->get_text(), &color );
		}

		// drawing
		if ( renders[pane] != NULL )
			renders[pane]->draw();
	}
}
// processing function
void gui::Menu::process( int milliseconds ){
	if ( ! capturing ){
		// if we are not in the middle of a capturing process
		if ( ! on ){
			// if menu is off and someone presses escape: turn on
			if ( Input::get_agg_escape() )
				turnon();
		}else{
			// if menu is on, send events to the active pane
			if ( Input::get_agg_escape() ){
				// escape event
				panes[pane]->exit();
				if ( on )
					Audio::playSound( Audio::SOUND_MENU_EXIT );
			}else if ( Input::get_agg_move() != util::Vector(0.0, 0.0) ){
				time += milliseconds;
				if ( time >= rc::menu_response_time() ){
					// if a movement has been done, it can be either change or movement event
					time = 0;
					if ( std::abs(Input::get_agg_move().cx()) > std::abs(Input::get_agg_move().cy()) ){
						// change event
						panes[pane]->change( (Input::get_agg_move().cx()<0.0) ? -1 : 1 );
						Audio::playSound( Audio::SOUND_MENU_CHANGE );
					}else{
						// movement event
						panes[pane]->move( (Input::get_agg_move().cy()<0.0) ? 1 : -1 );
						Audio::playSound( Audio::SOUND_MENU_MOVE );
					}
				}
			}else if ( Input::get_agg_bomb() ){
				// accept event
				panes[pane]->accept();
				Audio::playSound( Audio::SOUND_MENU_ACCEPT );
			}else{
				// user is doing nothing, next immediate action doesn't need to be ignored
				time = rc::menu_response_time();
			}
		}
	}else{
		// if we are capturing the joystick axis, we may be done
		if ( capture_type == CAPTURE_JOYSTICK_AXIS ){
			// ignore data in first half (because previous movement may still affect us)
			if ( capture_jtime < 0 ){
				capture_jvalue = 0;
				capture_jcount = false;
			}

			// count time
			if ( (capture_jtime < 0) || capture_jcount )
				capture_jtime += milliseconds;

			// if time margin has elapsed, notify the capture
			if ( capture_jtime >= rc::joystick_capturing_time() ){
				panes[capture_notifywho]->notify( true, SDLK_UNKNOWN, capture_jwhich, capture_jaxis, (capture_jvalue<0) );
				capture_type = NO_CAPTURE;
			}
		}
		// if the capture has ended (or been canceled), we are no longer capturing
		if ( capture_type == NO_CAPTURE ){
			capturing = false;
			// let the possible movement be ignored
			time = 0;
		}
	}
}
