// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef NDEBUG
# define BOOST_SPIRIT_DEBUG
#endif
#include <cmath>
#include <cstdlib>
#include <set>
#include <map>
#include <iostream>
#include <boost/spirit/iterator/file_iterator.hpp>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#define int_drawingcmds_t int,drawingcmds_t
#include "../vis/Draw.hpp"
#include "../Main.hpp"
#include "Grammar.hpp"
#include "Constants.hpp"

// where the constants are stored
#define BOOL_VAR( x ) static bool x##_var;
#define INT_VAR( x )  static int x##_var;
#define REAL_VAR( x ) static real x##_var;
#include "Macro.hpp"
#undef BOOL_VAR
#undef INT_VAR
#undef REAL_VAR

// getters for the constants
#define BOOL_VAR( x ) \
	bool rc::x(){ \
		return x##_var; \
	}
#define INT_VAR( x ) \
	int rc::x(){ \
		return x##_var; \
	}
#define REAL_VAR( x ) \
	real rc::x(){ \
		return x##_var; \
	}
#include "Macro.hpp"
#undef BOOL_VAR
#undef INT_VAR
#undef REAL_VAR

#if !defined(RC_FILE_PREFIX)
# define RC_FILE_PREFIX FILES_DIRECTORY
#endif // not defined RC_FILE_PREFIX

// rc file name
#define RC_NAME PACKAGE_TARNAME ".rc"

// some debugging macros
#ifndef NDEBUG
# define DEBUG_FORMAL_PARAM       , const std::string& _i
# define DEBUG_REAL_PARAM         , std::string()
# define DEBUG_DEEPEN_PARAM       , _i + "| "
# define DEBUG_DEEPEN2_PARAM      , _i + "| | "
# define DEBUG_SAME_PARAM         , _i
# define DEBUG_PRESENT_NODE(n,t)  std::cout << _i << to_string(n) << " (" << (t) << " " << to_id(n) << ")" << std::endl
# define DEBUG_PRESENT2_NODE(n,t) {std::cout << "| "; DEBUG_PRESENT_NODE(n,t);}
# define DEBUG_CHILD_NODE(n,t,i)  DEBUG_PRESENT2_NODE((n).children[i],t)
#else
# define DEBUG_FORMAL_PARAM
# define DEBUG_REAL_PARAM
# define DEBUG_DEEPEN_PARAM
# define DEBUG_DEEPEN2_PARAM
# define DEBUG_SAME_PARAM
# define DEBUG_PRESENT_NODE(n,t)
# define DEBUG_PRESENT2_NODE(n,t)
# define DEBUG_CHILD_NODE(n,t,i)
#endif

// identifiers for drawing operations
enum drawing_ops {
	THIN_LINES,
	NORMAL_LINES,
	COLOR,
	TRANSLATE,
	ROTATE,
	SCALE,
	SQUARE,
	CONCAVE_SQUARE,
	RHOMB,
	TRIANGLE,
	CROSS_PLUS,
	CIRCLE,
	CIRCULAR_SAW,
	SOLID_QUADS,
	LINE_LOOP,
	LINE_STRIP,
	LINES
};

// some useful ast types
typedef boost::spirit::file_iterator<>                             iterator_t;
typedef boost::spirit::tree_parse_info<iterator_t>                 result_t;
typedef boost::spirit::tree_match<iterator_t>::node_t              tnode_t;
typedef boost::spirit::tree_match<iterator_t>::const_tree_iterator titerator_t;
// some useful data types
typedef std::pair< std::vector<real>, std::vector<bool> > drawingparams_t;
typedef std::pair< int, drawingparams_t >                 drawingcmd_t;
typedef std::list< drawingcmd_t >                         drawingcmds_t;

// all variables defined in the rc
static std::map<std::string, double> definitions; // unordered
// holds all drawing definitions
static std::map<int, drawingcmds_t> drawings;
// the parse has been successful?
static bool errors = false;

// some useful functions
inline static std::size_t to_id( const tnode_t& n ){
	return n.value.id().to_long();
}
inline static std::string to_string( const tnode_t& n ){
	return std::string(n.value.begin(), n.value.end());
}
inline static char to_char( const tnode_t& n ){
	return *n.value.begin();
}
inline static bool is_integer( const double x ){
	return std::floor(x) == x;
}

// assigns the temporal values to its static place
static void bool_assign( const char *name, bool& bvar ){
	std::map<std::string, double>::const_iterator iter = definitions.find( name );
	if ( iter != definitions.end() )
		bvar = (*iter).second != 0;
	else{
		std::cerr << "ERROR: variable \"" << name << "\" must be defined in the rc." << std::endl;
		errors = true;
	}
}
static void int_assign( const char *name, int& ivar ){
	std::map<std::string, double>::const_iterator iter = definitions.find( name );
	if ( iter != definitions.end() )
		ivar = static_cast<int>( (*iter).second );
	else{
		std::cerr << "ERROR: variable \"" << name << "\" must be defined in the rc." << std::endl;
		errors = true;
	}
}
static void real_assign( const char *name, real& rvar ){
	std::map<std::string, double>::const_iterator iter = definitions.find( name );
	if ( iter != definitions.end() )
		rvar = static_cast<real>( (*iter).second );
	else{
		std::cerr << "ERROR: variable \"" << name << "\" must be defined in the rc." << std::endl;
		errors = true;
	}
}
static void assign(){
#define BOOL_VAR( x ) bool_assign( #x, x##_var );
#define INT_VAR( x )  int_assign( #x, x##_var );
#define REAL_VAR( x ) real_assign( #x, x##_var );
#include "Macro.hpp"
#undef BOOL_VAR
#undef INT_VAR
#undef REAL_VAR

	if ( errors ){
		std::cerr << "FATAL: the rc does not define all the needed variables." << std::endl;
		Main::quit( Main::ERROR_UNDEFINED_RC );
	}
}

// checks that constraints are satisfied
static void constraints(){
	const char *constraint__integer[] = { "width", "height",
		"colliderpassive_resolution", "blast_t", "examinis_t", "lagvard_t",
		"lightning_t", "lightning_dead_t", "player_imm_t", "player_reb_t",
		"redhole_t", "scared_t", "spawn_t", "spawnship_t", "bullet_lifes",
		"examinis_apicula_lifes", "examinis_queen_lifes", "laser_lifes",
		"lightning_lifes", "lightning_main_lifes", "redhole_lifes",
		"shielded_lifes", "examinis_hit_score", "examinis_kill_score",
		"lagvard_kill_score", "prob_variation_int", "proton_kill_score",
		"redhole_hit_score", "redhole_kill_score", "scared_kill_score",
		"shielded_hit_score", "shielded_kill_score", "spawned_kill_score",
		"spawnship_kill_score", "debris_milliseconds_ago",
		"player_defense_blinking_time", "player_defense_blink_time",
		"player_litter_int", "rubble_milliseconds_ago", "normal_shots_int",
		"fast_shots_int", "multi_shots_int", "laser_shots_int",
		"lightning_shots_int", "normal_shots_rate", "fast_shots_rate",
		"multi_shots_rate", "laser_shots_rate", "lightning_shots_rate",
		"multi_shots_amount", "powerup_lifes", "powerup_bombs",
		"enemy_threshold", "peace_time", "examinis_time", "lagvard_time",
		"redhole_time", "scared_time", "shielded_time", "spawnship_time",
		"examinis_int", "lagvard_int", "redhole_int", "scared_int",
		"shielded_int", "spawnship_int", "menu_response_time",
		"joystick_capturing_time", "gravity_int", "pursuit_int",
		"dodging_int", "collide_int", "font_point_size" };
	const char *constraint__gt_zero[] = { "width", "height",
		"colliderpassive_resolution", "lightning_dead_t", "bullet_lifes",
		"examinis_apicula_lifes", "examinis_queen_lifes", "laser_lifes",
		"lightning_lifes", "lightning_main_lifes", "redhole_lifes",
		"shielded_lifes", "redhole_mass", "spring_mass", "powerup_bombs",
		"examinis_int", "lagvard_int", "redhole_int", "scared_int",
		"shielded_int", "spawnship_int", "menu_response_time",
		"joystick_capturing_time", "font_point_size" };
	const char *constraint__ge_zero__sum_gt_zero[] = { "normal_shots_int",
		"fast_shots_int", "multi_shots_int", "laser_shots_int",
		"lightning_shots_int" };
	const char *constraint__ge_zero[] = { "powerup_lifes", "lagvard_mass",
		"scared_mass", "spawned_mass", "spawnship_mass", "proton_proportion",
		"examinis_prob", "examinis_prob_variation", "lagvard_prob",
		"lagvard_prob_variation", "redhole_prob", "redhole_prob_variation",
		"scared_prob", "scared_prob_variation", "shielded_prob",
		"shielded_prob_variation", "spawnship_prob", "spawnship_prob_variation",
		"no_spawning_prob", "no_spawning_prob_variation" };
	const char *constraint__gt_one[] = { "multi_shots_amount" };
	const char *constraint__neq_zero[] = { "examinis_orbitmin_r",
		"shielded_inner_r", "spawn_t", "player_defense_blink_time",
		"spring_color_magnitude" };

	foreach( const char *i, constraint__integer ){
		if ( ! is_integer( definitions[i] ) ){
			std::cerr << "ERROR: variable " << i << " must have an integer value." << std::endl;
			errors = true;
		}
	}
	foreach( const char *i, constraint__gt_zero ){
		if ( ! (definitions[i] > 0) ){
			std::cerr << "ERROR: variable " << i << " must be greater than 0." << std::endl;
			errors = true;
		}
	}
	double sum = 0;
	foreach( const char *i, constraint__ge_zero__sum_gt_zero ){
		if ( ! (definitions[i] >= 0) ){
			std::cerr << "ERROR: variable " << i << " must be greater than or equal to 0." << std::endl;
			errors = true;
		}
		sum += definitions[i];
	}
	if ( ! (sum > 0) ){
		std::cerr << "ERROR: variables";
		foreach( const char *i, constraint__ge_zero__sum_gt_zero )
			std::cerr << " " << i;
		std::cerr << " must be greater than 0 when added." << std::endl;
		errors = true;
	}
	foreach( const char *i, constraint__ge_zero ){
		if ( ! (definitions[i] >= 0) ){
			std::cerr << "ERROR: variable " << i << " must be greater than or equal to 0." << std::endl;
			errors = true;
		}
	}
	foreach( const char *i, constraint__gt_one ){
		if ( ! (definitions[i] > 1) ){
			std::cerr << "ERROR: variable " << i << " must be greater than 1." << std::endl;
			errors = true;
		}
	}
	foreach( const char *i, constraint__neq_zero ){
		if ( ! (definitions[i] != 0) ){
			std::cerr << "ERROR: variable " << i << " must not be 0." << std::endl;
			errors = true;
		}
	}

	if ( errors ){
		std::cerr << "FATAL: the rc defines invalid values." << std::endl;
		Main::quit( Main::ERROR_CONSTRAINTS_RC );
	}
}

// executes an arithmetic expression
static double expr( const tnode_t& e DEBUG_FORMAL_PARAM ){
	std::string aux;
	double one;
	double two;
	std::map<std::string, double>::const_iterator iter;

	switch( to_id(e) ){
		case boost::spirit::runcommands::EXPR_ID:
			// an ADD or SUB
			DEBUG_PRESENT_NODE(e,"add_sub");
			one = expr( e.children[0] DEBUG_DEEPEN_PARAM );
			two = expr( e.children[1] DEBUG_DEEPEN_PARAM );
			switch( to_char(e) ){
				case '+':
					return one + two;
				case '-':
					return one - two;
				default:
					throw std::string("ERROR: unknown operation ") + to_char(e) + ".";
			}
		case boost::spirit::runcommands::EXPR1_ID:
			// a MUL, DIV or MOD
			DEBUG_PRESENT_NODE(e,"mul_div_mod");
			one = expr( e.children[0] DEBUG_DEEPEN_PARAM );
			two = expr( e.children[1] DEBUG_DEEPEN_PARAM );
			switch( to_char(e) ){
				case '*':
					return one * two;
				case '/':
					if ( two != 0 )
						return one / two;
					else
						throw std::string("ERROR: division by 0.");
				case '%':
					if ( two != 0 )
						return std::fmod( one, two );
					else
						throw std::string("ERROR: module by 0.");
				default:
					throw std::string("ERROR: unknown operation ") + to_char(e) + ".";
			}
		case boost::spirit::runcommands::EXPR2_ID:
			// a POW
			DEBUG_PRESENT_NODE(e,"pow");
			one = expr( e.children[0] DEBUG_DEEPEN_PARAM );
			two = expr( e.children[1] DEBUG_DEEPEN_PARAM );
			if ( (one >= 0) || is_integer(two) )
				return std::pow( one, two );
			else
				throw std::string("ERROR: power would be a complex number.");
		case boost::spirit::runcommands::EXPR3_ID:
			// an unary PLUS or MIN
			DEBUG_PRESENT_NODE(e,"uplus_umin");
			one = expr( e.children[0] DEBUG_DEEPEN_PARAM );
			switch( to_char(e) ){
				case '+':
					return one;
				case '-':
					return -one;
				default:
					throw std::string("ERROR: unknown operation ") + to_char(e) + ".";
			}
		case boost::spirit::runcommands::EXPR4_ID:
			// a UREAL literal
			DEBUG_PRESENT_NODE(e,"ureal");
			aux = to_string(e);
			return std::strtod(aux.c_str(), NULL);
		case boost::spirit::runcommands::IDENT_ID:
			// an IDENT
			DEBUG_PRESENT_NODE(e,"ident");
			iter = definitions.find(to_string(e));
			if ( iter != definitions.end() )
				return (*iter).second;
			else
				throw std::string("ERROR: undefined variable ") + to_string(e) + ".";
		case boost::spirit::runcommands::FUNC_NAME_ID:
			// a FUNCTION CALL
			DEBUG_PRESENT_NODE(e,"func");
			one = expr( e.children[0] DEBUG_DEEPEN_PARAM );
			aux = to_string(e);
			if ( aux == "floor" ){
				return std::floor( one );
			}else if ( aux == "ceil" ){
				return std::ceil( one );
			}else if ( aux == "sin" ){
				return std::sin( one );
			}else if ( aux == "cos" ){
				return std::cos( one );
			}else if ( aux == "tan" ){
				return std::tan( one );
			}else if ( aux == "sqrt" ){
				if ( one >= 0 )
					return std::sqrt( one );
				else
					throw std::string("ERROR: sqrt would be a complex number.");
			}else if ( aux == "abs" ){
				return std::abs( one );
			}else if ( aux == "ln" ){
				if ( one > 0 )
					return std::log( one );
				else
					throw std::string("ERROR: ln needs a positive value.");
			}else
				throw std::string("ERROR: unknown function ") + aux + ".";
		default:
			throw std::string("ERROR: unknown operation ") + to_string(e) + ".";
	}
}
// executes drawing elements
static drawingcmds_t elements( titerator_t iter, const titerator_t end DEBUG_FORMAL_PARAM ){
	drawingcmds_t res;

	// for all elements
	for( ; iter != end ; ++iter ){
		DEBUG_PRESENT_NODE(*iter,"elem");

		// treating each type
		std::string aux = to_string( *iter );
		if ( aux == "thin_lines" ){
			res.push_back( drawingcmd_t( THIN_LINES, drawingparams_t( std::vector<real>(), std::vector<bool>() ) ) );
		}else if ( aux == "normal_lines" ){
			res.push_back( drawingcmd_t( NORMAL_LINES, drawingparams_t( std::vector<real>(), std::vector<bool>() ) ) );
		}else if ( aux == "color" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[1] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[2] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[3] DEBUG_DEEPEN_PARAM ) ) );
			foreach( real r, comps ){
				if ( (r < 0) || (r > 1) )
					throw std::string("ERROR: colors must be defined in the range [0,1].");
			}
			res.push_back( drawingcmd_t( COLOR, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "translate" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0].children[0] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[0].children[1] DEBUG_DEEPEN_PARAM ) ) );
			res.push_back( drawingcmd_t( TRANSLATE, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "rotate" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			res.push_back( drawingcmd_t( ROTATE, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "scale" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			if ( (*iter).children.size() == 2 )
				comps.push_back( static_cast<real>( expr( (*iter).children[1] DEBUG_DEEPEN_PARAM ) ) );
			else
				comps.push_back( comps.front() );
			res.push_back( drawingcmd_t( SCALE, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "square" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			res.push_back( drawingcmd_t( SQUARE, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "concave_square" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[1] DEBUG_DEEPEN_PARAM ) ) );
			res.push_back( drawingcmd_t( CONCAVE_SQUARE, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "rhomb" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			res.push_back( drawingcmd_t( RHOMB, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "triangle" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			res.push_back( drawingcmd_t( TRIANGLE, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "cross_plus" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			res.push_back( drawingcmd_t( CROSS_PLUS, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "circle" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[1] DEBUG_DEEPEN_PARAM ) ) );
			if ( ! is_integer(comps.back()) )
				throw std::string("ERROR: circle's points must be an integer value.");
			if ( comps.back() <= 2 )
				throw std::string("ERROR: circle's points must be greater than 2.");
			res.push_back( drawingcmd_t( CIRCLE, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( aux == "circular_saw" ){
			std::vector<real> comps;
			comps.push_back( static_cast<real>( expr( (*iter).children[0] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[1] DEBUG_DEEPEN_PARAM ) ) );
			comps.push_back( static_cast<real>( expr( (*iter).children[2] DEBUG_DEEPEN_PARAM ) ) );
			if ( ! is_integer(comps.back()) )
				throw std::string("ERROR: circular_saw's teeth must be an integer value.");
			if ( comps.back() <= 2 )
				throw std::string("ERROR: circular_saw's teeth must be greater than 2.");
			res.push_back( drawingcmd_t( CIRCULAR_SAW, drawingparams_t( comps, std::vector<bool>() ) ) );
		}else if ( (aux == "solid_quads") || (aux == "line_loop") || (aux == "line_strip") || (aux == "lines") ){
			// saving id, the rest is equal for all of them
			int id = 0;
			if ( aux == "solid_quads" ){
				id = SOLID_QUADS;
			}else if ( aux == "line_loop" ){
				id = LINE_LOOP;
			}else if ( aux == "line_strip" ){
				id = LINE_STRIP;
			}else if ( aux == "lines" ){
				id = LINES;
			}
			// saving points
			std::vector<real> comps;
			std::vector<bool> speci;
			foreach( const tnode_t& n, (*iter).children ){
				switch( to_id(n) ){
					case boost::spirit::runcommands::VERTEX_ID:
						DEBUG_PRESENT2_NODE(n,"point");
						comps.push_back( static_cast<real>( expr( n.children[0] DEBUG_DEEPEN2_PARAM ) ) );
						comps.push_back( static_cast<real>( expr( n.children[1] DEBUG_DEEPEN2_PARAM ) ) );
						speci.push_back( false );
						break;
					case boost::spirit::runcommands::CURVED_VERTEX_ID:
						DEBUG_PRESENT2_NODE(n,"curved_point");
						comps.push_back( static_cast<real>( expr( n.children[0].children[0] DEBUG_DEEPEN2_PARAM ) ) );
						comps.push_back( static_cast<real>( expr( n.children[0].children[1] DEBUG_DEEPEN2_PARAM ) ) );
						comps.push_back( static_cast<real>( expr( n.children[1].children[0] DEBUG_DEEPEN2_PARAM ) ) );
						comps.push_back( static_cast<real>( expr( n.children[1].children[1] DEBUG_DEEPEN2_PARAM ) ) );
						comps.push_back( static_cast<real>( expr( n.children[2].children[0] DEBUG_DEEPEN2_PARAM ) ) );
						comps.push_back( static_cast<real>( expr( n.children[2].children[1] DEBUG_DEEPEN2_PARAM ) ) );
						speci.push_back( true );
						break;
					default:
						throw std::string("ERROR: unknown vertex point for freeform drawings.");
				}
			}
			res.push_back( drawingcmd_t( id, drawingparams_t( comps, speci ) ) );
		}else
			throw std::string("ERROR: unknown drawing operation ") + aux + ".";
	}

	return res;
}
// executes a condition
static bool cond( const tnode_t& c DEBUG_FORMAL_PARAM ){
	std::string aux;
	double one;
	double two;

	switch( to_id(c) ){
		case boost::spirit::runcommands::COND_ID:
			// an AND or OR
			DEBUG_PRESENT_NODE(c,"and_or");
			switch( to_char(c) ){
				case '&':
					return cond(c.children[0] DEBUG_DEEPEN_PARAM) && cond(c.children[1] DEBUG_DEEPEN_PARAM);
				case '|':
					return cond(c.children[0] DEBUG_DEEPEN_PARAM) || cond(c.children[1] DEBUG_DEEPEN_PARAM);
				default:
					throw std::string("ERROR: unknown conditional operation ") + to_string(c) + ".";
			}
		case boost::spirit::runcommands::COND1_ID:
			// a NOT
			DEBUG_PRESENT_NODE(c,"not");
			switch( to_char(c) ){
				case '~':
					return ! cond(c.children[0] DEBUG_DEEPEN_PARAM);
				default:
					throw std::string("ERROR: unknown conditional operation ") + to_string(c) + ".";
			}
		case boost::spirit::runcommands::COND2_ID:
			// a def(IDENT)
			DEBUG_PRESENT_NODE(c,"def");
			DEBUG_CHILD_NODE(c,"ident",0);
			return definitions.find(to_string(c.children[0])) != definitions.end();
		case boost::spirit::runcommands::COMP_ID:
			// a comparison
			DEBUG_PRESENT_NODE(c,"comp");
			one = expr( c.children[0] DEBUG_DEEPEN_PARAM );
			two = expr( c.children[1] DEBUG_DEEPEN_PARAM );
			aux = to_string(c);
			if ( aux == "<=" )
				return one <= two;
			else if ( aux == ">=" )
				return one >= two;
			else if ( aux == "==" )
				return one == two;
			else if ( aux == "<" )
				return one < two;
			else if ( aux == ">" )
				return one > two;
			else
				throw std::string("ERROR: unknown comparison operation ") + aux + ".";
		default:
			throw std::string("ERROR: unknown conditional operation ") + to_string(c) + ".";
	}
}
// prototype
static void execute( titerator_t iter, const titerator_t& end DEBUG_FORMAL_PARAM );
// executes a statement
static void stmt( const tnode_t& s DEBUG_FORMAL_PARAM ){
	double aux;
	std::string nam;
	int id;

	switch( to_id(s) ){
		case boost::spirit::runcommands::ASSIGN_ID:
			// an ASSIGN
			DEBUG_PRESENT_NODE(s,"assig");
			DEBUG_CHILD_NODE(s,"ident",0);
			aux = expr( s.children[1] DEBUG_DEEPEN_PARAM );
			if ( std::isfinite(aux) )
				definitions[to_string(s.children[0])] = aux;
			else
				throw std::string("ERROR: trying to assign a non finite value to ") + to_string(s.children[0]) + ".";
			break;
		case boost::spirit::runcommands::FIGURE_NAME_ID:
			// a FIGURE
			DEBUG_PRESENT_NODE(s,"figure");

			// translating the text to an id
			nam = to_string(s);
			if ( nam == "bullet" )
				id = vis::BULLET_ID;
			else if ( nam == "examinis_apicula" )
				id = vis::EXAMINIS_APICULA_ID;
			else if ( nam == "examinis_queen" )
				id = vis::EXAMINIS_QUEEN_ID;
			else if ( nam == "lagvard" )
				id = vis::LAGVARD_ID;
			else if ( nam == "player_defense" )
				id = vis::PLAYER_ID0;
			else if ( nam == "player" )
				id = vis::PLAYER_ID;
			else if ( nam == "proton" )
				id = vis::PROTON_ID;
			else if ( nam == "redhole" )
				id = vis::REDHOLE_ID;
			else if ( nam == "scared" )
				id = vis::SCARED_ID;
			else if ( nam == "shielded" )
				id = vis::SHIELDED_ID;
			else if ( nam == "spawned" )
				id = vis::SPAWNED_ID;
			else if ( nam == "spawnship" )
				id = vis::SPAWNSHIP_ID;
			else
				throw std::string("ERROR: unknown figure name ") + nam + ".";

			// storing the information
			drawings.insert( std::pair<int,drawingcmds_t>( id, elements( s.children.begin(), s.children.end() DEBUG_DEEPEN_PARAM ) ) );
			break;
		case boost::spirit::runcommands::CONDITIONAL_ID:
			// a CONDITIONAL
			DEBUG_PRESENT_NODE(s,"if");
			if ( cond( s.children[0] DEBUG_DEEPEN_PARAM ) ){
				DEBUG_CHILD_NODE(s,"then",1);
				execute( s.children[1].children.begin(), s.children[1].children.end() DEBUG_DEEPEN2_PARAM );
			}else if ( s.children.size() == 3 ){
				DEBUG_CHILD_NODE(s,"else",2);
				execute( s.children[2].children.begin(), s.children[2].children.end() DEBUG_DEEPEN2_PARAM );
			}
			break;
		default:
			throw std::string("ERROR: unknown statement ") + to_string(s) + ".";
	}
}
// executes a list of statements
static void execute( titerator_t iter, const titerator_t& end DEBUG_FORMAL_PARAM ){
	for( ; iter != end ; ++iter ){
		try{
			stmt( *iter DEBUG_SAME_PARAM );
		}catch( const std::string& err ){
			std::cerr << err << std::endl;
			errors = true;
		}
	}
}

// sets up a new definition
void rc::add_define( const std::string& d ){
	std::string::size_type p = d.find_first_of('=');
	if ( p == std::string::npos )
		definitions[d] = 0;
	else if ( p > 0 )
		definitions[ d.substr(0,p) ] = (p==d.size()-1) ? 0 : std::strtod(d.substr(p+1).c_str(), NULL);
}
// initializes the rc [depends on: -]
void rc::init(){
	// file's begin and end
	std::string filename;
	iterator_t first;
	iterator_t last;
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
	filename = RC_NAME;
	first = iterator_t( filename );
#else // Windows
	char *home = std::getenv( "HOME" );
	if ( home != NULL ){
		filename = home;
		filename += "/." RC_NAME;
		first = iterator_t( filename );
		if ( ! first ){
			filename = RC_FILE_PREFIX RC_NAME;
			first = iterator_t( filename );
		}
	}else{
		filename = RC_FILE_PREFIX RC_NAME;
		first = iterator_t( filename );
	}
#endif // Not Windows
	if ( ! first ){
		std::cerr << "FATAL: unable to open rc file \"" << filename << "\"." << std::endl;
		Main::quit( Main::ERROR_LOAD_RC );
	}
	last = first.make_end();

	// parsing file
	boost::spirit::runcommands g;
	boost::spirit::skipper s;
	result_t data = ast_parse( first, last, g, s );
	if ( ! data.match ){
		std::cerr << "FATAL: could not properly parse the rc file \"" << filename << "\"." << std::endl;
		Main::quit( Main::ERROR_PARSE_RC );
	}

	// execute the rc code
	execute( (*data.trees.begin()).children.begin(), (*data.trees.begin()).children.end() DEBUG_REAL_PARAM );
	if ( errors ){
		std::cerr << "FATAL: errors while executing the rc file \"" << filename << "\"." << std::endl;
		Main::quit( Main::ERROR_EXEC_RC );
	}

	// storing
	assign();

	// checking
	constraints();

	// cleaning data
	definitions.clear();
}
// initialize all recorded drawings
void rc::initalldrawings(){
	int j;

	// for all definied figures
	foreach( const std::pair<int_drawingcmds_t>& d, drawings ){
		vis::Draw::start_record( d.first );
		// for all its elements
		foreach( const drawingcmd_t& c, d.second ){
			const std::vector<real>& p = c.second.first;
			const std::vector<bool>& b = c.second.second;
			switch( c.first ){
				case THIN_LINES:
					vis::Draw::set_thin_lines();
					break;
				case NORMAL_LINES:
					vis::Draw::set_normal_lines();
					break;
				case COLOR:
					vis::Draw::set_color( p[0], p[1], p[2], p[3] );
					break;
				case TRANSLATE:
					vis::Draw::translate( util::Vector(p[0], p[1]) );
					break;
				case ROTATE:
					vis::Draw::rotate( p[0] );
					break;
				case SCALE:
					vis::Draw::scale( p[0], p[1] );
					break;
				case SQUARE:
					vis::Draw::square( p[0] );
					break;
				case CONCAVE_SQUARE:
					vis::Draw::concave_square( p[0], p[1] );
					break;
				case RHOMB:
					vis::Draw::rhomb( p[0] );
					break;
				case TRIANGLE:
					vis::Draw::triangle( p[0] );
					break;
				case CROSS_PLUS:
					vis::Draw::cross_plus( p[0] );
					break;
				case CIRCLE:
					vis::Draw::circle( p[0], static_cast<int>(p[1]) );
					break;
				case CIRCULAR_SAW:
					vis::Draw::circular_saw( p[0], p[1], static_cast<int>(p[2]) );
					break;
				case SOLID_QUADS:
				case LINE_LOOP:
				case LINE_STRIP:
				case LINES:
					// starting up the context
					switch( c.first ){
						case SOLID_QUADS:
							vis::Draw::start_solid_quads();
							break;
						case LINE_LOOP:
							vis::Draw::start_line_loop();
							break;
						case LINE_STRIP:
							vis::Draw::start_line_strip();
							break;
						case LINES:
							vis::Draw::start_lines();
							break;
					}
					// drawing
					j = 0;
					for( int i=0 ; i<static_cast<int>(b.size()) ; ++i ){
						if ( b[i] ){
							vis::Draw::add_curve( p[j], p[j+1], p[j+2], p[j+3], p[j+4], p[j+5] );
							j += 6;
						}else{
							vis::Draw::add_point( p[j], p[j+1] );
							j += 2;
						}
					}
					// ending the context
					switch( c.first ){
						case SOLID_QUADS:
							vis::Draw::end_solid_quads();
							break;
						case LINE_LOOP:
							vis::Draw::end_line_loop();
							break;
						case LINE_STRIP:
							vis::Draw::end_line_strip();
							break;
						case LINES:
							vis::Draw::end_lines();
							break;
					}
					break;
			}
		}
		vis::Draw::end_record();
	}
}
// ends the rc
void rc::finalize(){
	definitions.clear();
	drawings.clear();
}
