// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef GRAMMAR_HPP
#define GRAMMAR_HPP

#include <boost/spirit/core.hpp>
#include <boost/spirit/tree/ast.hpp>
#include <boost/spirit/utility/confix.hpp>

namespace boost { namespace spirit {

// Grammar for useful data. It recognizes arithmetic expresions, drawings and
// some control flow.
struct runcommands : public grammar<runcommands>{
	enum rule_id {
		S_ID = 1,
		STMT_ID,
		ASSIGN_ID,
		EXPR_ID,
		EXPR1_ID, // 5
		EXPR2_ID,
		EXPR3_ID,
		EXPR4_ID,
		IDENT_ID,
		FUNC_NAME_ID, // 10
		PARAM_ID,
		FIGURE_ID,
		FIGURE_NAME_ID,
		ELEMENT_ID,
		DRAWING_OPERATION_ID, // 15
		DRAWING_FORM_ID,
		DRAWING_FREEFORM_ID,
		VERTEX_ID,
		CURVED_VERTEX_ID,
		CONDITIONAL_ID, // 20
		COND_ID,
		COND1_ID,
		COND2_ID,
		COMP_ID,
		THEN_BODY_ID, // 25
		ELSE_BODY_ID
	};

	template<typename ScannerT>
	struct definition{
		definition( const runcommands& self ){
			S = *( stmt ) >> end_p;

			stmt
				= assign
				| figure
				| conditional
				;

			assign
				= ident >> root_node_d[ch_p('=')] >> expr
				;
			ident
				= token_node_d[lexeme_d[(ch_p('_') | alpha_p) >> *(ch_p('_') | alnum_p)]
				               - func_name
				               - figure_name
				               - str_p("if")]
				;
			expr
				= expr1 >> *( root_node_d[ch_p('+') | ch_p('-')] >> expr1 )
				;
			expr1
				= expr2 >> *( root_node_d[ch_p('*') | ch_p('/') | ch_p('%')] >> expr2 )
				;
			expr2
				= expr3 >> !( root_node_d[ch_p('^')] >> expr2 )
				;
			expr3
				= !( root_node_d[ch_p('+') | ch_p('-')] ) >> expr4
				;
			expr4
				= ident
				| root_node_d[func_name] >> param
				| ureal_p
				| inner_node_d[ch_p('(') >> expr >> ch_p(')')]
				;
			func_name
				= str_p("floor")
				| str_p("ceil")
				| str_p("sin")
				| str_p("cos")
				| str_p("tan")
				| str_p("sqrt")
				| str_p("abs")
				| str_p("ln")
				;
			param
				= inner_node_d[ch_p('(') >> expr >> ch_p(')')]
				;

			figure
				= root_node_d[figure_name] >> no_node_d[ch_p('{')]
						>> *( inner_node_d[ch_p('{') >> element >> ch_p('}')] )
					>> no_node_d[ch_p('}')]
				;
			figure_name
				= str_p("bullet")
				| str_p("examinis_apicula")
				| str_p("examinis_queen")
				| str_p("lagvard")
				| str_p("player_defense")
				| str_p("player")
				| str_p("proton")
				| str_p("redhole")
				| str_p("scared")
				| str_p("shielded")
				| str_p("spawned")
				| str_p("spawnship")
				;
			element
				= drawing_operation
				| drawing_form
				| drawing_freeform
				;
			drawing_operation
				= root_node_d[str_p("thin_lines")]
				| root_node_d[str_p("normal_lines")]
				| root_node_d[str_p("color")] >> no_node_d[str_p("red") >> str_p("=")] >> expr
				                              >> no_node_d[str_p("green") >> str_p("=")] >> expr
				                              >> no_node_d[str_p("blue") >> str_p("=")] >> expr
				                              >> no_node_d[str_p("alpha") >> str_p("=")] >> expr
				| root_node_d[str_p("translate")] >> vertex
				| root_node_d[str_p("rotate")] >> expr
				| root_node_d[str_p("scale")] >> expr >> !( expr )
				;
			drawing_form
				= root_node_d[str_p("square")] >> no_node_d[str_p("side") >> str_p("=")] >> expr
				| root_node_d[str_p("concave_square")] >> no_node_d[str_p("side") >> str_p("=")] >> expr
				                                       >> no_node_d[str_p("inner") >> str_p("=")] >> expr
				| root_node_d[str_p("rhomb")] >> no_node_d[str_p("diagonal") >> str_p("=")] >> expr
				| root_node_d[str_p("triangle")] >> no_node_d[str_p("radius") >> str_p("=")] >> expr
				| root_node_d[str_p("cross_plus")] >> no_node_d[str_p("length") >> str_p("=")] >> expr
				| root_node_d[str_p("circle")] >> no_node_d[str_p("radius") >> str_p("=")] >> expr
				                               >> no_node_d[str_p("points") >> str_p("=")] >> expr
				| root_node_d[str_p("circular_saw")] >> no_node_d[str_p("inner_radius") >> str_p("=")] >> expr
				                                     >> no_node_d[str_p("outter_radius") >> str_p("=")] >> expr
				                                     >> no_node_d[str_p("teeth") >> str_p("=")] >> expr
				;
			drawing_freeform
				= root_node_d[str_p("solid_quads")] >> *( vertex )
				| root_node_d[str_p("line_loop")] >> *( vertex | curved_vertex )
				| root_node_d[str_p("line_strip")] >> *( vertex | curved_vertex )
				| root_node_d[str_p("lines")] >> *( vertex )
				;
			vertex
				= no_node_d[ch_p('(')] >> expr >> no_node_d[ch_p(',')] >> expr >> no_node_d[ch_p(')')]
				;
			curved_vertex
				= no_node_d[ch_p('{')] >> vertex >> vertex >> vertex >> no_node_d[ch_p('}')]
				;

			conditional
				= root_node_d[str_p("if")] >> cond >> then_body >> !( else_body ) >> no_node_d[str_p("endif")]
				;
			cond
				= cond1 >> *( root_node_d[ch_p('|') | ch_p('&')] >> cond1 )
				;
			cond1
				= !( root_node_d[ch_p('~')] ) >> cond2
				;
			cond2
				= root_node_d[str_p("def")] >> inner_node_d[ch_p('(') >> ident >> ch_p(')')]
				| inner_node_d[ch_p('(') >> cond >> ch_p(')')]
				| comp
				;
			comp
				= expr >> root_node_d[str_p("<=") | str_p(">=") | str_p("==") | ch_p('<') | ch_p('>')] >> expr
				;
			then_body
				= root_node_d[str_p("then")] >> *( stmt )
				;
			else_body
				= root_node_d[str_p("else")] >> *( stmt )
				;

			BOOST_SPIRIT_DEBUG_NODE(S);
			BOOST_SPIRIT_DEBUG_NODE(stmt);
			BOOST_SPIRIT_DEBUG_NODE(assign);
			BOOST_SPIRIT_DEBUG_NODE(expr);
			BOOST_SPIRIT_DEBUG_NODE(expr1);
			BOOST_SPIRIT_DEBUG_NODE(expr2);
			BOOST_SPIRIT_DEBUG_NODE(expr3);
			BOOST_SPIRIT_DEBUG_NODE(expr4);
			BOOST_SPIRIT_DEBUG_NODE(ident);
			BOOST_SPIRIT_DEBUG_NODE(func_name);
			BOOST_SPIRIT_DEBUG_NODE(param);
			BOOST_SPIRIT_DEBUG_NODE(figure);
			BOOST_SPIRIT_DEBUG_NODE(figure_name);
			BOOST_SPIRIT_DEBUG_NODE(element);
			BOOST_SPIRIT_DEBUG_NODE(drawing_operation);
			BOOST_SPIRIT_DEBUG_NODE(drawing_form);
			BOOST_SPIRIT_DEBUG_NODE(drawing_freeform);
			BOOST_SPIRIT_DEBUG_NODE(vertex);
			BOOST_SPIRIT_DEBUG_NODE(curved_vertex);
			BOOST_SPIRIT_DEBUG_NODE(conditional);
			BOOST_SPIRIT_DEBUG_NODE(cond);
			BOOST_SPIRIT_DEBUG_NODE(cond1);
			BOOST_SPIRIT_DEBUG_NODE(cond2);
			BOOST_SPIRIT_DEBUG_NODE(comp);
			BOOST_SPIRIT_DEBUG_NODE(then_body);
			BOOST_SPIRIT_DEBUG_NODE(else_body);
		}

		rule< ScannerT, parser_context<>, parser_tag<S_ID> >                 S;
		rule< ScannerT, parser_context<>, parser_tag<STMT_ID> >              stmt;
		rule< ScannerT, parser_context<>, parser_tag<ASSIGN_ID> >            assign;
		rule< ScannerT, parser_context<>, parser_tag<EXPR_ID> >              expr;
		rule< ScannerT, parser_context<>, parser_tag<EXPR1_ID> >             expr1;
		rule< ScannerT, parser_context<>, parser_tag<EXPR2_ID> >             expr2;
		rule< ScannerT, parser_context<>, parser_tag<EXPR3_ID> >             expr3;
		rule< ScannerT, parser_context<>, parser_tag<EXPR4_ID> >             expr4;
		rule< ScannerT, parser_context<>, parser_tag<IDENT_ID> >             ident;
		rule< ScannerT, parser_context<>, parser_tag<FUNC_NAME_ID> >         func_name;
		rule< ScannerT, parser_context<>, parser_tag<PARAM_ID> >             param;
		rule< ScannerT, parser_context<>, parser_tag<FIGURE_ID> >            figure;
		rule< ScannerT, parser_context<>, parser_tag<FIGURE_NAME_ID> >       figure_name;
		rule< ScannerT, parser_context<>, parser_tag<ELEMENT_ID> >           element;
		rule< ScannerT, parser_context<>, parser_tag<DRAWING_OPERATION_ID> > drawing_operation;
		rule< ScannerT, parser_context<>, parser_tag<DRAWING_FORM_ID> >      drawing_form;
		rule< ScannerT, parser_context<>, parser_tag<DRAWING_FREEFORM_ID> >  drawing_freeform;
		rule< ScannerT, parser_context<>, parser_tag<VERTEX_ID> >            vertex;
		rule< ScannerT, parser_context<>, parser_tag<CURVED_VERTEX_ID> >     curved_vertex;
		rule< ScannerT, parser_context<>, parser_tag<CONDITIONAL_ID> >       conditional;
		rule< ScannerT, parser_context<>, parser_tag<COND_ID> >              cond;
		rule< ScannerT, parser_context<>, parser_tag<COND1_ID> >             cond1;
		rule< ScannerT, parser_context<>, parser_tag<COND2_ID> >             cond2;
		rule< ScannerT, parser_context<>, parser_tag<COMP_ID> >              comp;
		rule< ScannerT, parser_context<>, parser_tag<THEN_BODY_ID> >         then_body;
		rule< ScannerT, parser_context<>, parser_tag<ELSE_BODY_ID> >         else_body;

		const rule< ScannerT, parser_context<>, parser_tag<S_ID> >& start() const{
			return S;
		}
	};
};

// Grammar for skipped data: space and comments (C and C++) are skipped.
struct skipper : public grammar<skipper>{
	template<typename ScannerT>
	struct definition{
		definition( const skipper& self ){
			skip
				= no_node_d[space_p]
				| no_node_d[comment_p("//")]
				| no_node_d[comment_p("/*","*/")]
				;

			BOOST_SPIRIT_DEBUG_NODE(skip);
		}

		rule< ScannerT, parser_context<> > skip;

		const rule< ScannerT, parser_context<> >& start() const{
			return skip;
		}
	};
};

} }

#endif
