// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#ifndef PARTITION_HPP
#define PARTITION_HPP

#include <cmath>
#include <set>
#include <list>
#include <vector>
#include <utility>
#include <algorithm>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "../rc/Constants.hpp"
#include "Matrix.hpp"
#include "Vector.hpp"
#ifndef NDEBUG
# include "../vis/Draw.hpp"
#endif

namespace util {

template<typename T>
class Partition{
	private:
		// type of the buckets' indexes
		typedef std::pair<int,int> bucketind_t;

	public:
		// ctor for a ring partition of roughly partitions*partitions buckets
		Partition( int partitions, int dimensions ) :
		 k( partitions/rc::width() ),
		 c( 1.0/k ),
		 d(),
		 rows( std::max(1, static_cast<int>(std::floor(rc::height()*k))) ),
		 cols( partitions ),
		 elements(0)
		{
			for( ; dimensions ; --dimensions )
				d.push_back( new Matrix< std::set<T> >( rows, cols ) );
		}
		// dtor
		virtual ~Partition(){
			for( int t=0 ; t<static_cast<int>(d.size()) ; ++t )
				delete d[t];
			d.clear();
		}

		// inserts a new element into the partition
		void insert( T element, int dimensions, const Vector& xy, real r ){
			// getting buckets' indexes
			std::list<bucketind_t> ind;
			compute_indexes( xy, r, ind );
			// inserting it
			insert( element, dimensions, ind );

			++elements;
		}
		// moves the element from its old position to the new one
		void update( T element, int dimensions, const Vector& old_xy, real old_r, const Vector& new_xy, real new_r ){
			// getting old buckets' indexes
			std::list<bucketind_t> old_ind;
			compute_indexes( old_xy, old_r, old_ind );
			// getting new buckets' indexes
			std::list<bucketind_t> new_ind;
			compute_indexes( new_xy, new_r, new_ind );

			// sorting
			old_ind.sort();
			new_ind.sort();

			// places where to erase the element
			std::list<bucketind_t> to_erase( old_ind.size() );
			std::list<bucketind_t>::iterator it_to_erase;
			it_to_erase = std::set_difference( old_ind.begin(), old_ind.end(), new_ind.begin(), new_ind.end(), to_erase.begin() );
			to_erase.erase( it_to_erase, to_erase.end() );
			// places where to insert the element
			std::list<bucketind_t> to_insert( new_ind.size() );
			std::list<bucketind_t>::iterator it_to_insert;
			it_to_insert = std::set_difference( new_ind.begin(), new_ind.end(), old_ind.begin(), old_ind.end(), to_insert.begin() );
			to_insert.erase( it_to_insert, to_insert.end() );

			// updating
			erase( element, dimensions, to_erase );
			insert( element, dimensions, to_insert );
		}
		// erases the element (whose last parameters were xy and r)
		void erase( T element, int dimensions, const Vector& xy, real r ){
			// getting buckets' indexes
			std::list<bucketind_t> ind;
			compute_indexes( xy, r, ind );
			// erasing it
			erase( element, dimensions, ind );

			--elements;
		}

		// gets the elements that are in the specified circle
		void candidates( int dimensions, const Vector& xy, real r, std::list<T>& res ) const{
			// getting buckets' indexes
			std::list<bucketind_t> ind;
			compute_indexes( xy, r, ind );
			// unioning all elements
			foreach( const bucketind_t& bi, ind ){
				for( int t=0 ; (dimensions>>t) && (t<static_cast<int>(d.size())) ; ++t ){
					if ( ! ((dimensions>>t) & 1) )
						continue;

					res.insert( res.end(),
					            (*d[t])[bi.first][bi.second].begin(),
					            (*d[t])[bi.first][bi.second].end() );
				}
			}
			// cleaning
			res.sort();
			res.unique();
		}

		// tests if there are any elements
		bool empty() const{
			return elements == 0;
		}

#ifndef NDEBUG
		// paints the buckets with elements
		void draw(){
			real colors_r[] = {1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.5, 0.5, 1.0, 0.5};
			real colors_g[] = {0.5, 0.5, 1.0, 1.0, 0.0, 0.0, 0.5, 0.5, 0.5, 1.0};
			real colors_b[] = {0.0, 1.0, 0.5, 0.0, 1.0, 0.0, 0.5, 1.0, 0.5, 0.5};
			real colors_a[] = {0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5};
			for( int t=0 ; t<static_cast<int>(d.size()) ; ++t ){
				vis::Draw::set_color( colors_r[t], colors_g[t], colors_b[t], colors_a[t] );
				for( int j=0 ; j<rows ; ++j ){
					for( int i=0 ; i<cols ; ++i ){
						if ( (*d[t])[j][i].empty() )
							continue;

						vis::Draw::start_solid_quads();
							vis::Draw::add_point( i*c,     j*c );
							vis::Draw::add_point( (i+1)*c, j*c );
							vis::Draw::add_point( (i+1)*c, (j+1)*c );
							vis::Draw::add_point( i*c,     (j+1)*c );
						vis::Draw::end_solid_quads();
					}
				}
			}
		}
#endif

	private:
		// saves in res the indexes of the buckets that are included in the circle centered at xy with radius r
		void compute_indexes( const Vector& xy, real r, std::list<bucketind_t>& res ) const{
			// translates the coordinate to its corresponding bucket indexes
			int i = std::min( cols-1, std::max( 0, static_cast<int>(std::floor(xy.cx()*k)) ) );
			int j = std::min( rows-1, std::max( 0, static_cast<int>(std::floor(xy.cy()*k)) ) );

			// adding exact bucket
			res.push_back( bucketind_t(j,i) );
			// moving to the left
			for( int ii = i-1 ; (ii >= 0) && ((ii+1)*c > xy.cx()-r) ; --ii )
				res.push_back( bucketind_t(j,ii) );
			// moving to the right
			for( int ii = i+1 ; (ii < cols) && (ii*c < xy.cx()+r) ; ++ii )
				res.push_back( bucketind_t(j,ii) );

			real r2 = r * r; // using squares to avoid sqrts
			// moving up
			for( int jj = j+1 ; (jj < rows) && (jj*c < xy.cy()+r) ; ++jj ){
				res.push_back( bucketind_t(jj,i) );

				// and then moving left
				for( int ii = i-1 ; (ii >= 0) && ((Vector((ii+1)*c,jj*c)-xy).magnitude2() < r2) ; --ii )
					res.push_back( bucketind_t(jj,ii) );
				// and then moving right
				for( int ii = i+1 ; (ii < cols) && ((Vector(ii*c,jj*c)-xy).magnitude2() < r2) ; ++ii )
					res.push_back( bucketind_t(jj,ii) );
			}
			// moving down
			for( int jj = j-1 ; (jj >= 0) && ((jj+1)*c > xy.cy()-r) ; --jj ){
				res.push_back( bucketind_t(jj,i) );

				// and then moving left
				for( int ii = i-1 ; (ii >= 0) && ((Vector((ii+1)*c,(jj+1)*c)-xy).magnitude2() < r2) ; --ii )
					res.push_back( bucketind_t(jj,ii) );
				// and then moving right
				for( int ii = i+1 ; (ii < cols) && ((Vector(ii*c,(jj+1)*c)-xy).magnitude2() < r2) ; ++ii )
					res.push_back( bucketind_t(jj,ii) );
			}
		}
		// inserts the element in all indexes
		void insert( T element, int dimensions, const std::list<bucketind_t>& ind ){
			foreach( const bucketind_t& bi, ind ){
				for( int t=0 ; (dimensions>>t) && (t<static_cast<int>(d.size())) ; ++t ){
					if ( ! ((dimensions>>t) & 1) )
						continue;
					(*d[t])[bi.first][bi.second].insert( element );
				}
			}
		}
		// erases the element from all indexes
		void erase( T element, int dimensions, const std::list<bucketind_t>& ind ){
			foreach( const bucketind_t& bi, ind ){
				for( int t=0 ; (dimensions>>t) && (t<static_cast<int>(d.size())) ; ++t ){
					if ( ! ((dimensions>>t) & 1) )
						continue;
					(*d[t])[bi.first][bi.second].erase( element );
				}
			}
		}

		// multiplicative constant to translate a coordinate to a bucket index
		real k;
		// multiplicative constant to translate a bucket index to a coordinate
		real c;
		// matrix vector that stores the elements
		std::vector< Matrix< std::set<T> >* > d; // unordered
		// matrix dimensions
		int rows;
		int cols;
		// number of elements present in this partition
		int elements;
};

}

#endif // Partition.hpp
