// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <cmath>
#include <algorithm>
#include <SDL_opengl.h>
#include "../rc/Constants.hpp"
#include "Screen.hpp"
#include "Draw.hpp"

// variables to remember some parameters
real vis::Draw::previous_color[] = {0.0, 0.0, 0.0, 0.0};
real vis::Draw::present_color[]  = {0.0, 0.0, 0.0, 0.0};
real vis::Draw::present_point[]  = {0.0, 0.0};

// record operations that will be performed
void vis::Draw::start_record( int id ){
	glNewList( id, GL_COMPILE );
}
void vis::Draw::end_record(){
	glEndList();
}
void vis::Draw::play_record( int id ){
	glCallList( id );
}

// drawing line's width
void vis::Draw::set_thin_lines(){
	glLineWidth( rc::thin_line_width() );
}
void vis::Draw::set_normal_lines(){
	glLineWidth( rc::normal_line_width() );
}

// free solid-quads drawing context
void vis::Draw::start_solid_quads(){
	glBegin( GL_QUADS );
}
void vis::Draw::end_solid_quads(){
	glEnd();
}
// free line-loop drawing context
void vis::Draw::start_line_loop(){
	glBegin( GL_LINE_LOOP );
}
void vis::Draw::end_line_loop(){
	glEnd();
}
// free line-strip drawing context
void vis::Draw::start_line_strip(){
	glBegin( GL_LINE_STRIP );
}
void vis::Draw::end_line_strip(){
	glEnd();
}
// free lines drawing context
void vis::Draw::start_lines(){
	glBegin( GL_LINES );
}
void vis::Draw::end_lines(){
	glEnd();
}
// draws a square
void vis::Draw::square( real side ){
	real s2 = side/2.0;
	start_line_loop();
		add_point(  s2,  s2 );
		add_point( -s2,  s2 );
		add_point( -s2, -s2 );
		add_point(  s2, -s2 );
	end_line_loop();
}
// draws a concave square
void vis::Draw::concave_square( real side, real inner ){
	real s2 = side/2.0;
	real i2 = inner/2.0;
	start_line_loop();
		add_point( s2, 0.0 );
		add_curve(  i2, 0.0, 0.0,  i2, 0.0,  s2 );
		add_curve( 0.0,  i2, -i2, 0.0, -s2, 0.0 );
		add_curve( -i2, 0.0, 0.0, -i2, 0.0, -s2 );
		add_curve( 0.0, -i2,  i2, 0.0,  s2, 0.0 );
	end_line_loop();
}
// draws a rhomb
void vis::Draw::rhomb( real diagonal ){
	real d2 = diagonal/2.0;
	start_line_loop();
		add_point(  d2, 0.0 );
		add_point( 0.0,  d2 );
		add_point( -d2, 0.0 );
		add_point( 0.0, -d2 );
	end_line_loop();
}
// draws a triangle
void vis::Draw::triangle( real radius ){
	real x2 = radius*std::cos( PI/2.0 - PI/3.0 );
	real y2 = radius*std::sin( PI/2.0 - PI/3.0 );
	start_line_loop();
		add_point( 0.0, radius );
		add_point( -x2,    -y2 );
		add_point(  x2,    -y2 );
	end_line_loop();
}
// draws a plus
void vis::Draw::cross_plus( real length ){
	real l2 = length/2.0;
	start_lines();
		add_point(  l2, 0.0 );
		add_point( -l2, 0.0 );
		add_point( 0.0,  l2 );
		add_point( 0.0, -l2 );
	end_lines();
}
// draws a circle (points > 2)
void vis::Draw::circle( real radius, int points ){
	start_line_loop();
		for( real i=0.0 ; i<PI*2.0 ; i+=(PI*2.0)/points )
			add_point( radius*std::cos(i), radius*std::sin(i) );
	end_line_loop();
}
// draws a circular saw (teeth > 2)
void vis::Draw::circular_saw( real inner_r, real outter_r, int teeth ){
	start_line_loop();
		for( real i=0.0 ; i<PI*2.0 ; i+=(PI*2.0)/teeth ){
			add_point( outter_r*std::cos(i), outter_r*std::sin(i) );
			add_point(  inner_r*std::cos(i),  inner_r*std::sin(i) );
		}
	end_line_loop();
}

// adds a point to the current path
void vis::Draw::add_point( const util::Vector& p ){
	add_point( p.cx(), p.cy() );
}
void vis::Draw::add_point( real x, real y ){
	present_point[0] = x;
	present_point[1] = y;
	glVertex2( x, y );
}
// adds a curve to the current path from previous point to (x3,y3)
// with (x1,y1) and (x2,y2) as control points, without changing color
void vis::Draw::add_curve( real x1, real y1, real x2, real y2, real x3, real y3 ){
	add_curve( util::Vector(x1,y1), util::Vector(x2,y2), util::Vector(x3,y3), false );
}
// adds a curve to the current path from previous point to (x3,y3)
// with (x1,y1) and (x2,y2) as control points, with a color gradient
void vis::Draw::add_curve( const util::Vector& p1, real x2, real y2, real x3, real y3 ){
	add_curve( p1, util::Vector(x2,y2), util::Vector(x3,y3), true );
}
void vis::Draw::add_curve( real x1, real y1, const util::Vector& p2, const util::Vector& p3 ){
	add_curve( util::Vector(x1,y1), p2, p3, true );
}
void vis::Draw::add_curve( const util::Vector& p1, const util::Vector& p2, const util::Vector& p3 ){
	add_curve( p1, p2, p3, true );
}
void vis::Draw::add_curve( const util::Vector& p1, const util::Vector& p2, const util::Vector& p3, bool color ){
	// shortcut: if the curve is invisible no computed points are needed
	if ( (previous_color[3]<rc::visible_alpha_threshold()) && (present_color[3]<rc::visible_alpha_threshold()) ){
		add_point( p3 );
		return;
	}

	// path will go from p0 to p3 with p1 and p2 as control points
	util::Vector p0( present_point[0], present_point[1] );

	// heuristically computing the amount of points that will be drawn for this curve
	real magnitude = (p1-p0).magnitude() + (p2-p1).magnitude() + (p3-p2).magnitude();
	real guess = std::min( rc::bezier_max(), std::ceil(rc::bezier_heuristic()*magnitude*vis::Screen::get_pixelsize()) );
	int stops = std::max( 2, static_cast<int>(guess) );

	// color gradient
	real r = previous_color[0];
	real g = previous_color[1];
	real b = previous_color[2];
	real a = previous_color[3];
	real dr = (present_color[0]-r) / stops;
	real dg = (present_color[1]-g) / stops;
	real db = (present_color[2]-b) / stops;
	real da = (present_color[3]-a) / stops;

	// movement along the bezier curve
	real t  = 0.0;
	real dt = 1.0 / stops;

	// factors
	util::Vector fc = (p1 - p0)*3.0;
	util::Vector fb = (p2 - p1)*3.0 - fc;
	util::Vector fa = p3 - p0 - fc - fb;

	// drawing curve
	for( int i=0 ; i<stops ; ++i ){
		// advancing
		t += dt;
		// updating color
		if ( color ){
			r += dr;
			g += dg;
			b += db;
			a += da;
		}

		// current position
		real t2 = t  * t;
		real t3 = t2 * t;

		// present point
		util::Vector p = (fa * t3) + (fb * t2) + (fc * t) + p0;

		// drawing
		if ( color )
			set_color( r, g, b, a );
		add_point( p );
	}
}

// sets current color
void vis::Draw::set_color( real r, real g, real b, real a ){
	previous_color[0] = present_color[0];
	previous_color[1] = present_color[1];
	previous_color[2] = present_color[2];
	previous_color[3] = present_color[3];
	present_color[0] = r;
	present_color[1] = g;
	present_color[2] = b;
	present_color[3] = a;
	glColor4( r, g, b, a );
}

// clears current drawing state
void vis::Draw::reset(){
	glLoadIdentity();
	set_normal_lines();
	set_color( rc::clear_c_r(), rc::clear_c_g(), rc::clear_c_b(), rc::clear_c_a() );
}

// matrix manipulation
void vis::Draw::translate( const util::Vector& p ){
	glTranslate( p.cx(), p.cy(), 0.0 );
}
void vis::Draw::rotate( real angle ){
	glRotate( angle*(180.0/PI), 0.0, 0.0, 1.0 );
}
void vis::Draw::scale( real s ){
	glScale( s, s, 1.0 );
}
void vis::Draw::scale( real sx, real sy ){
	glScale( sx, sy, 1.0 );
}
