// nvsts is a multi-directional shooter game.
// Copyright (C) 2008 Carles Creus Lopez
//
// nvsts is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>
// or write to the Free Software Foundation, Inc., 51 Franklin Street,
// Fifth Floor, Boston, MA 02110-1301 USA.
//
// You may contact the author at <crazycraft@gmail.com>

#include <algorithm>
#include <boost/foreach.hpp>
#define foreach BOOST_FOREACH
#include "../rc/Constants.hpp"
#include "../vis/Draw.hpp"
#include "../Main.hpp"
#include "Blast.hpp"

// ctor
au::Blast::Blast( const util::Vector& xy, real r, int t, bool active ) :
 Automaton( xy, util::Vector(0.0,0.0), r, t ),
 rub(),
 active( active )
{
	if ( active )
		start_activecollider( BLAST_COL );
	start_processable();
	start_visible( RUBBLE_PRIO );
}
// dtor
au::Blast::~Blast(){
}
// each collision type for active colliders
void au::Blast::recieve_collision( int& res, Debris *p ){
	if ( ! p->intersects_circle( xy, r-rc::blast_w() ) ){ // acting as an annulus
		rubble_type r;
		r.xy = p->get_xy();
		r.v  = (r.xy - xy).magnitude(rc::blast_s());
		r.red   = p->get_red();
		r.green = p->get_green();
		r.blue  = p->get_blue();
		r.alpha = p->get_alpha();
		rub.push_back( r );
		res = PASSIVE_DIES;
	}else
		res = NO_DEATHS;
}
// serializes the automaton
void au::Blast::serialize( std::ostream& out ){
	out << Automaton::BLAST_SERIAL
	    << " " << xy << " " << r << " " << t
	    << " " << active << " " << rub.size();
	foreach( const rubble_type& r, rub ){
		out << " " << r.xy
		    << " " << r.v
		    << " " << r.red
		    << " " << r.green
		    << " " << r.blue
		    << " " << r.alpha;
	}
	out << std::endl;
}
// unserializes the automaton
void au::Blast::unserialize( std::istream& in ){
	util::Vector xy;
	real r;
	int t;
	bool active;
	int rs;

	in >> xy >> r >> t
	   >> active >> rs;
	if ( r < 0 )
		throw Main::UNSERIAL_BLAST;
	if ( t < 0 )
		throw Main::UNSERIAL_BLAST;
	if ( rs < 0 )
		throw Main::UNSERIAL_RUBBLE;

	Blast *b = new Blast( xy, r, t, active );

	for( ; rs ; --rs ){
		rubble_type r;
		in >> r.xy
		   >> r.v
		   >> r.red
		   >> r.green
		   >> r.blue
		   >> r.alpha;
		if ( (r.alpha < 0) || (r.alpha > 1) )
			throw Main::UNSERIAL_RUBBLE;
		b->rub.push_back( r );
	}
}
// drawing function
void au::Blast::specific_draw(){
	vis::Draw::set_thin_lines();
	vis::Draw::start_lines();
		// for each rubble
		foreach( const rubble_type& r, rub ){
			vis::Draw::set_color( r.red, r.green, r.blue, r.alpha );
			vis::Draw::add_point( r.xy - r.v*rc::rubble_milliseconds_ago() );
			vis::Draw::add_point( r.xy );
		}
	vis::Draw::end_lines();
}
// process function
void au::Blast::specific_process( int milliseconds ){
	// processing the blast
	if ( active && (t > rc::blast_t()) ){
		active = false;
		end_activecollider();
	}
	if ( active )
		r += rc::blast_s() * milliseconds;
	else if ( rub.empty() ){
		add_zombie();
		return;
	}

	// using this automaton to compute rubble bounces
	util::Vector tmp_xy = xy;
	util::Vector tmp_v  = v;
	real         tmp_r  = r;
	r = rc::rubble_r();
	// for each rubble
	for( std::list<rubble_type>::iterator iter=rub.begin() ; iter != rub.end() ; ){
		// moving
		(*iter).xy += (*iter).v*milliseconds;

		// bouncing with this automaton
		xy = (*iter).xy;
		v  = (*iter).v;
		if ( checkRing_bounce(1.0) ){
			(*iter).xy = xy;
			(*iter).v  = v;
		}

		// if blast is gone, the rubble is dying
		if ( ! active ){
			(*iter).alpha *= std::max( 0.0, 1.0-rc::rubble_fade_out()*milliseconds );
			if ( (*iter).alpha < rc::rubble_min_visiblity() )
				iter = rub.erase( iter );
			else{
				(*iter).v *= 1.0 - rc::rubble_s_lose()*milliseconds; // lineally decreasing speed
				++iter;
			}
		}else{
			(*iter).alpha = std::min( 1.0, (*iter).alpha*(1.0 + rc::rubble_fade_out()*milliseconds) );
			++iter;
		}
	}
	// restoring values
	xy = tmp_xy;
	v  = tmp_v;
	r  = tmp_r;
}
